"""Project settings. There is no need to edit this file unless you want to change values
from the Kedro defaults. For further information, including these default values, see
https://kedro.readthedocs.io/en/stable/kedro_project_setup/settings.html."""

# Instantiated project hooks.
from quafel.hooks import ProjectHooks, PipelineHooks, DataCatalogHooks
from kedro_viz.integrations.kedro.sqlite_store import SQLiteStore
from pathlib import Path

HOOKS = (ProjectHooks(), PipelineHooks(), DataCatalogHooks())

# Installed plugins for which to disable hook auto-registration.
# DISABLE_HOOKS_FOR_PLUGINS = ("kedro-viz",)

# Class that manages storing KedroSession data.
SESSION_STORE_CLASS = SQLiteStore
SESSION_STORE_ARGS = {"path": str(Path(__file__).parents[2] / "data")}

# Class that manages Kedro's library components.
# from kedro.framework.context import KedroContext
# CONTEXT_CLASS = KedroContext

# Directory that holds configuration.
# CONF_SOURCE = "conf"

# Class that manages how configuration is loaded.
# CONFIG_LOADER_CLASS = TemplatedConfigLoader
# # Keyword arguments to pass to the `CONFIG_LOADER_CLASS` constructor.
# CONFIG_LOADER_ARGS = {
#     "globals_pattern": "*data_science.yml",
# }

# Class that manages the Data Catalog.
# from kedro.io import DataCatalog
# DATA_CATALOG_CLASS = DataCatalog



from qiskit.circuit import (
    ClassicalRegister,
    QuantumCircuit,
    CircuitInstruction,
    ParameterVector,
    Reset,
)
from qiskit.circuit.library import standard_gates
from qiskit.circuit.exceptions import CircuitError
from qiskit.quantum_info import partial_trace, random_unitary
from qiskit import execute
from qiskit_aer import StatevectorSimulator
import numpy as np

import jax
import jax.numpy as jnp
from typing import List, Dict, Tuple
import pandas as pd
import logging
import os
import hashlib

log = logging.getLogger(__name__)


def log_circuit(qasm_circuit):
    return {"circuit_image": None}


def generate_evaluation_partitions(evaluation_matrix, skip_combinations):
    partitions = {}
    idx = 0
    for f in evaluation_matrix["frameworks"]:
        if "qubits" in skip_combinations:
            q = max(evaluation_matrix["qubits"])
            for d in evaluation_matrix["depths"]:
                for s in evaluation_matrix["shots"]:
                    partitions[f"{idx}"] = {
                        "framework": f,
                        "qubits": q,
                        "depth": d,
                        "shots": s,
                    }
                    idx += 1
        elif "depth" in skip_combinations:
            d = max(evaluation_matrix["depth"])
            for q in evaluation_matrix["qubits"]:
                for s in evaluation_matrix["shots"]:
                    partitions[f"{idx}"] = {
                        "framework": f,
                        "qubits": q,
                        "depth": d,
                        "shots": s,
                    }
                    idx += 1
        elif "shots" in skip_combinations:
            s = max(evaluation_matrix["shots"])
            for d in evaluation_matrix["depths"]:
                for q in evaluation_matrix["qubits"]:
                    partitions[f"{idx}"] = {
                        "framework": f,
                        "qubits": q,
                        "depth": d,
                        "shots": s,
                    }
                    idx += 1
        else:
            for q in evaluation_matrix["qubits"]:
                for d in evaluation_matrix["depths"]:
                    for s in evaluation_matrix["shots"]:
                        partitions[f"{idx}"] = {
                            "framework": f,
                            "qubits": q,
                            "depth": d,
                            "shots": s,
                        }
                        idx += 1

    # Add row index, because the order of the order
    # of the rows might get scrambled otherwise
    eval_partitions = pd.DataFrame(
        partitions, index=["framework", "qubits", "depth", "shots"]
    )
    log.info(f"Generated {eval_partitions.shape[1]} partitions")
    return {"evaluation_partitions": eval_partitions}


def extract_partition_data(partition):
    # TODO: improve this by accessing data by name
    framework = partition[partition.columns[0]][0]
    qubits = int(partition[partition.columns[0]][1])
    depth = int(partition[partition.columns[0]][2])
    shots = int(partition[partition.columns[0]][3])

    return {
        "qubits": qubits,
        "depth": depth,
        "n_shots": shots,
        "framework": framework,
    }


def qasm_circuit_to_qiskit(qasm_circuit):
    qc = QuantumCircuit.from_qasm_str(qasm_circuit)

    return {
        "qiskit_circuit": qc,
    }


def _random_circuit(
    num_qubits,
    depth,
    max_operands=2,
    conditional=False,
    reset=False,
    seed=None,
):
    """
    <<<<<<< HEAD
        Code partly taken from Qiskit:
        qiskit/circuit/random/utils.py

        Generates a random quantum circuit with the number of qubits
        and circuit depth provided.

        The generating method will favor multi-qubit gates.

        Parameterized gates will have UNBOUND parameters, thus it is
        required to bind them before the circuit can be evaluated.

        Args:
            num_qubits (int): Number of qubits.
            depth (int): Depth of the circuit.
            max_operands (int, optional): Maximum number of operands. Defaults to 2.
            conditional (bool, optional): Cond. Circuit?. Defaults to False.
            reset (bool, optional): Whether to reset the circuit. Defaults to False.
            seed (int, optional): Seed for random number generation. Defaults to None.

        Returns:
            QuantumCircuit: The generated quantum circuit.
    """
    if num_qubits == 0:
        return QuantumCircuit()
    if max_operands < 1 or max_operands > 4:
        raise CircuitError("max_operands must be between 1 and 4")
    max_operands = max_operands if num_qubits > max_operands else num_qubits

    gates_1q = [
        # (Gate class, number of qubits, number of parameters)
        (standard_gates.IGate, 1, 0),
        (standard_gates.XGate, 1, 0),
        (standard_gates.RZGate, 1, 1),
        (standard_gates.HGate, 1, 0),
        (standard_gates.RXGate, 1, 1),
        (standard_gates.RYGate, 1, 1),
        (standard_gates.SGate, 1, 0),
        (standard_gates.TGate, 1, 0),
        (standard_gates.U2Gate, 1, 2),
        (standard_gates.U3Gate, 1, 3),
        (standard_gates.YGate, 1, 0),
        (standard_gates.ZGate, 1, 0),
    ]
    if reset:
        gates_1q.append((Reset, 1, 0))
    gates_2q = [
        (standard_gates.CXGate, 2, 0),
        (standard_gates.CZGate, 2, 0),
        (standard_gates.SwapGate, 2, 0),
    ]
    gates_3q = [
        (standard_gates.CCXGate, 3, 0),
    ]

    # add gates to the overall set, depending on the number of operands
    gates = gates_1q.copy()
    if max_operands >= 2:
        gates.extend(gates_2q)
    if max_operands >= 3:
        gates.extend(gates_3q)
    gates = np.array(
        gates,
        dtype=[("class", object), ("num_qubits", np.int16), ("num_params", np.int32)],
    )

    # generate a numpy array, that we will use later, to fill "gaps"
    gates_1q = np.array(gates_1q, dtype=gates.dtype)

    qc = QuantumCircuit(num_qubits)

    cr = ClassicalRegister(num_qubits, "c")
    qc.add_register(cr)

    rng = np.random.default_rng(seed)

    qubits = np.array(qc.qubits, dtype=object, copy=True)

    # Apply arbitrary random operations in layers across all qubits.
    for layer_number in range(depth):
        # We generate all the randomness for the layer in one go,
        # to avoid many separate calls to
        # the randomisation routines, which can be fairly slow.

        # This reliably draws too much randomness,
        # but it's less expensive than looping over more
        # calls to the rng. After, trim it down by finding the point
        # when we've used all the qubits.
        gate_specs = rng.choice(gates, size=len(qubits))
        cumulative_qubits = np.cumsum(gate_specs["num_qubits"], dtype=np.int16)
        # Efficiently find the point in the list where the total gates
        # would use as many as
        # possible of, but not more than, the number of qubits in the layer.
        # If there's slack, fill
        # it with 1q gates.
        # This will, with favor multi qubit gates, but does not ensure >1 qubit gates
        # being used, especially with a low number of layers
        max_index = np.searchsorted(cumulative_qubits, num_qubits, side="right")
        gate_specs = gate_specs[:max_index]
        slack = num_qubits - cumulative_qubits[max_index - 1]
        if slack:
            gate_specs = np.hstack((gate_specs, rng.choice(gates_1q, size=slack)))

        # For efficiency in the Python loop, this uses Numpy vectorisation to
        # pre-calculate the
        # indices into the lists of qubits and parameters for every gate,
        # and then suitably
        # randomises those lists.
        q_indices = np.empty(len(gate_specs) + 1, dtype=np.int16)
        p_indices = np.empty(len(gate_specs) + 1, dtype=np.int16)
        q_indices[0] = p_indices[0] = 0
        np.cumsum(gate_specs["num_qubits"], out=q_indices[1:])
        np.cumsum(gate_specs["num_params"], out=p_indices[1:])
        # parameters = rng.uniform(0, 2 * np.pi, size=p_indices[-1])
        parameters = ParameterVector(f"p_{layer_number}", p_indices[-1])
        rng.shuffle(qubits)

        # We've now generated everything we're going to need.
        # Now just to add everything.  The
        # conditional check is outside the two loops to make
        # the more common case of no conditionals
        # faster, since in Python we don't have a compiler to do this for us.
        if conditional and layer_number != 0:
            is_conditional = rng.random(size=len(gate_specs)) < 0.1
            condition_values = rng.integers(
                0, 1 << min(num_qubits, 63), size=np.count_nonzero(is_conditional)
            )
            c_ptr = 0
            for gate, q_start, q_end, p_start, p_end, is_cond in zip(
                gate_specs["class"],
                q_indices[:-1],
                q_indices[1:],
                p_indices[:-1],
                p_indices[1:],
                is_conditional,
            ):
                operation = gate(*parameters[p_start:p_end])
                if is_cond:
                    qc.measure(qc.qubits, cr)
                    # The condition values are required to be bigints,
                    # not Numpy's fixed-width type.
                    operation.condition = (cr, int(condition_values[c_ptr]))
                    c_ptr += 1
                qc._append(
                    CircuitInstruction(
                        operation=operation, qubits=qubits[q_start:q_end]
                    )
                )
        else:
            for gate, q_start, q_end, p_start, p_end in zip(
                gate_specs["class"],
                q_indices[:-1],
                q_indices[1:],
                p_indices[:-1],
                p_indices[1:],
            ):
                operation = gate(*parameters[p_start:p_end])
                qc._append(
                    CircuitInstruction(
                        operation=operation, qubits=qubits[q_start:q_end]
                    )
                )

    return qc


def generate_random_qasm_circuit_from_partition(partition, seed):
    partition_data = extract_partition_data(partition)
    return {
        **generate_random_qasm_circuit(
            partition_data["qubits"], partition_data["depth"], seed
        ),
        "n_shots": partition_data["depth"],
        "framework": partition_data["framework"],
    }


def generate_random_qasm_circuit(
    qubits: int, depth: int, seed: int
) -> Dict[str, List[float]]:
    """
    Generate a random quantum circuit and its representation as a QASM string.
    Note that the QASM circuit differs from the original circuit in such a way
    that all qubits are being measured.

    Args:
        qubits: Number of qubits.
        depth: Number of gate layers.
        seed: Random seed.

    Returns:
        A dictionary with the key 'qasm_circuit' containing the QASM string and
        the key 'circuit' containing the Qiskit circuit object.
    """
    qc = _random_circuit(qubits, depth, max_operands=2, seed=seed)

    rng = np.random.default_rng(seed)

    # generate all the parameters of the circuit in one go
    parameter_values = rng.uniform(0, 2 * np.pi, size=len(qc.parameters))

    # bind the parameters to the circuit
    bound_circuit = qc.assign_parameters(
        {p: v for p, v in zip(qc.parameters, parameter_values)}
    )
    # measure all of the bound circuit
    # note that we explicitly NOT use measure_all because this would
    # add a barrier in the qasm representation that is not recognized by some
    # frameworks when translating back later
    bound_circuit.measure(bound_circuit.qubits, *bound_circuit.cregs)

    # return the bound circuit and the parameterizable circuit
    return {"qasm_circuit": bound_circuit.qasm(), "qiskit_circuit": qc}


def generate_evaluation_matrix(
    min_qubits: int,
    max_qubits: int,
    qubits_increment: int,
    qubits_type: int,
    min_depth: int,
    max_depth: int,
    depth_increment: int,
    depth_type: int,
    min_shots: int,
    max_shots: int,
    shots_increment: int,
    shots_type: int,
    frameworks: List[str],
):
    def generate_ticks(min_t, max_t, inc_t, type_t="linear"):
        if type_t == "linear":
            ticks = [i for i in range(min_t, max_t + inc_t, inc_t)]
        elif "exp" in type_t:
            base = int(type_t.split("exp")[1])
            ticks = [base**i for i in range(min_t, max_t + inc_t, inc_t)]
        else:
            raise ValueError("Unknown base specified and type is not linear")

        return ticks

    qubits = generate_ticks(min_qubits, max_qubits, qubits_increment, qubits_type)
    depths = generate_ticks(min_depth, max_depth, depth_increment, depth_type)
    shots = generate_ticks(min_shots, max_shots, shots_increment, shots_type)

    frameworks = frameworks

    return {
        "evaluation_matrix": {
            "frameworks": frameworks,
            "qubits": qubits,
            "depths": depths,
            "shots": shots,
        }
    }


def get_pqc_statevector(
    circuit: QuantumCircuit, params: np.ndarray, backend, precision: int, cache=True
) -> float:
    # Note that this is a jax rng, so it does not matter if we call that multiple times
    bound_circuit = circuit.bind_parameters(params)

    # the qasm representation contains the bound parameters, thus it is ok to hash that
    hs = hashlib.md5(bound_circuit.qasm().encode()).hexdigest()

    if cache:
        name = f"pqc_{hs}.npy"

        cache_folder = ".cache"
        if not os.path.exists(cache_folder):
            os.mkdir(cache_folder)

        file_path = os.path.join(cache_folder, name)

        if os.path.isfile(file_path):
            log.debug(f"Loading PQC statevector from {file_path}")
            loaded_array = np.load(file_path)
            return loaded_array

    # execute the PQC circuit with the current set of parameters
    # ansatz = circuit(params, circuit.num_qubits)
    result = execute(bound_circuit, backend=backend).result()

    # extract the statevector from the simulation result
    U = result.get_statevector(bound_circuit, decimals=precision).data

    if cache:
        log.debug(f"Cached PQC statevector into {file_path}")
        np.save(file_path, U)

    return U


def calculate_entangling_capability(
    circuit: QuantumCircuit,
    samples_per_parameter: int,
    samples_per_qubit: int,
    seed: int,
) -> Dict[str, float]:
    """
    Calculate the entangling capability of a quantum circuit.
    The strategy is taken from https://doi.org/10.48550/arXiv.1905.10876
    Implementation inspiration from
    https://obliviateandsurrender.github.io/blogs/expr.html

    Sanity Check; The following circuit should yield an entangling capability of 1.
    qc = QuantumCircuit(2)
    qc.h(0)
    qc.cx(0,1)
    qc.rx(*ParameterVector("x", 1), 0)

    Note that if qubits are being measured, this will return almost zero
    because the statevector simulator cannot work on collapsed states.

    If the circuit doesn't contain parameterizable operations, nan is returned.

    Args:
        circuit (QuantumCircuit): The quantum circuit.
        samples_per_parameter (int): The number of samples to generate.
        seed (int): The seed for the random number generator.

    Returns:
        dict: A dictionary containing the entangling capability value.
    """

    def meyer_wallach(circuit, samples, params_shape, precision, rng):
        if circuit.num_qubits == 1:
            return 0

        mw_measure = np.zeros(samples)

        # FIXME: unify the range for parameters in the circuit
        # generation method and the sampling here
        params = rng.uniform(0, 2 * np.pi, (samples, params_shape))

        # generate a list from [0..num_qubits-1]
        # we need that later to trace out the corresponding qubits
        qb = list(range(circuit.num_qubits))

        backend = StatevectorSimulator(precision="single")
        backend.set_options(
            max_parallel_threads=10,
            max_parallel_experiments=0,
            statevector_parallel_threshold=16,
        )

        # outer sum of the MW measure; iterate over set of parameters
        for i in range(samples):
            U = get_pqc_statevector(circuit, params[i], backend, precision)

            # generate a list from [0..num_qubits-1]
            # we need that later to trace out the corresponding qubits
            qb = list(range(circuit.num_qubits))
            # initialize the inner sum which corresponds to the entropy
            entropy = 0

            # inner sum of the MW measure
            for j in range(circuit.num_qubits):
                # density of the jth qubit after tracing out the rest
                density = partial_trace(U, qb[:j] + qb[j + 1 :]).data
                # trace of the density matrix
                entropy += np.trace(density @ density).real

            # fixes accumulating decimals that would otherwise lead to a MW > 1
            entropy = entropy / circuit.num_qubits
            # inverse of the normalized entropy is the MW
            # for the current sample of parameters
            mw_measure[i] = 1 - entropy
        # final normalization according to formula
        return max(0, min((2 * np.sum(mw_measure) / samples), 1))

    rng = np.random.default_rng(seed=seed)

    if len(circuit.parameters) == 0:
        entangling_capability = 0
    else:
        # TODO: propagate precision to kedro parameters
        entangling_capability = meyer_wallach(
            circuit=circuit,
            samples=samples_per_parameter * int(np.log(len(circuit.parameters)) + 1)
            + samples_per_qubit * circuit.num_qubits,
            params_shape=len(circuit.parameters),
            precision=5,
            rng=rng,
        )

    return {"entangling_capability": entangling_capability}


def calculate_expressibility(
    circuit: QuantumCircuit,
    samples_per_parameter: int,
    haar_samples_per_qubit: int,
    seed: int,
) -> Dict[str, float]:
    """
    Calculate the expressibility of a PQC circuit using
    a randomized estimation scheme.
    The strategy is taken from https://doi.org/10.48550/arXiv.1905.10876
    Implementation inspiration from
    https://obliviateandsurrender.github.io/blogs/expr.html

    Args:
        circuit (QuantumCircuit): The PQC circuit to be analyzed
        samples_per_parameter (int): The number of samples to use for estimation
        seed (int): The seed for the random number generator

    Returns:
        Dict[str, float]: A dictionary containing the expressibility value
    """

    def random_haar_unitary(n_qubits: int, rng: np.random.RandomState) -> np.ndarray:
        """
        Generate a random unitary matrix in the Haar measure.
        For details on the QR decomposition, see
        https://doi.org/10.48550/arXiv.math-ph/0609050

        Args:
            n_qubits (int): The number of qubits in the system
            rng (np.random.RandomState): The RandomState object to use for
                generating random numbers

        Returns:
            np.ndarray: A 2^n x 2^n unitary matrix representing a random
                unitary in the Haar measure on the unitary group U(2^n)
        """

        N = 2**n_qubits

        # # Generate uniformly sampled random complex numbers
        # Z = jax.random.normal(rng, (N, N)) + 1.0j * jax.random.normal(rng, (N, N))

        # def f(Z):
        #     # Do a QR decomposition
        #     [Q, R] = jnp.linalg.qr(Z)
        #     # Allow D following unitary matrix constraints
        #     D = jnp.diag(jnp.diagonal(R) / jnp.abs(jnp.diagonal(R)))
        #     # Composite the Haar Unitary
        #     return jnp.dot(Q, D)
        # return jax.jit(f)(Z)

        return random_unitary(
            N,
            int(rng._base_array.real[0]),
        ).data

    def haar_integral(
        n_qubits: int, samples: int, rng: np.random.RandomState
    ) -> np.ndarray:
        """
        Compute the Haar integral for a given number of samples

        Args:
            n_qubits (int): The number of qubits in the system
            samples (int): The number of samples to use for estimation
            rng (np.random.RandomState): The RandomState object to use for
                generating random numbers

        Returns:
            np.ndarray: A 2^n x 2^n array representing the normalized max
                expressibility
        """
        N = 2**n_qubits
        Z = jnp.zeros((N, N), dtype=complex)

        zero_state = jnp.zeros(N, dtype=complex)
        zero_state = zero_state.at[0].set(1)

        def f(zero_state, U):
            A = jnp.matmul(zero_state, U).reshape(-1, 1)
            return jnp.kron(A, A.conj().T)

        jf = jax.jit(f)

        for i in range(samples):
            rng, subkey = jax.random.split(rng)
            U = random_haar_unitary(n_qubits, subkey)

            Z += jf(zero_state, U)

        return Z / samples

    def get_haar_integral(
        n_qubits: int, samples: int, rng: np.random.RandomState, cache=True
    ) -> float:
        if cache:
            name = f"haar_{n_qubits}q_{samples}s.npy"

            cache_folder = ".cache"
            if not os.path.exists(cache_folder):
                os.mkdir(cache_folder)

            file_path = os.path.join(cache_folder, name)

            if os.path.isfile(file_path):
                log.debug(f"Loading Haar integral from {file_path}")
                loaded_array = np.load(file_path)
                return loaded_array

        # Note that this is a jax rng, so it does not matter if we
        # call that multiple times
        array = haar_integral(n_qubits, samples, rng)

        if cache:
            log.debug(f"Cached Haar integral into {file_path}")
            np.save(file_path, array)

        return array

    def pqc_integral(
        circuit: QuantumCircuit,
        samples: int,
        params_shape: Tuple,
        precision: int,
        rng: np.random.default_rng,
    ) -> np.ndarray:
        """
        Compute the entangling capability of a PQC circuit using a randomized
        estimation scheme

        Args:
            circuit (QuantumCircuit): The PQC circuit to be analyzed
            samples (int): The number of samples to use for estimation
            params_shape (tuple): The shape of the array of parameters to be
                used in the circuit simulation
            precision (int): The number of decimals to use when extracting the
                statevector from the simulation result

        Returns:
            np.ndarray: A 2^n x 2^n array representing the expressibility
            of the PQC circuit, where n is the number of qubits in the circuit
        """
        N = 2**circuit.num_qubits
        Z = jnp.zeros((N, N), dtype=complex)

        def f(U):
            return jnp.kron(U, U.conj().T)  # type: ignore

        jf = jax.jit(f)

        backend = StatevectorSimulator(precision="single")
        backend.set_options(
            max_parallel_threads=10,
            max_parallel_experiments=0,
            statevector_parallel_threshold=16,
        )

        # FIXME: unify the range for parameters in the circuit
        # generation method and the sampling here
        params = rng.uniform(0, 2 * np.pi, (samples, params_shape))
        for i in range(samples):
            # extract the statevector from the simulation result
            U = get_pqc_statevector(circuit, params[i], backend, precision).reshape(
                -1, 1
            )

            # accumulate the contribution to the expressibility
            Z += jf(U)

        return Z / samples

    rng = np.random.default_rng(seed=seed)
    jrng = jax.random.key(seed)

    def f(haar, pqc):
        # Note that we use the INVERSE here, because a
        # a LOW distance would actually correspond to a HIGH expressibility
        return 1 - jnp.linalg.norm(haar - pqc)

    jf = jax.jit(f)

    # FIXME: the actual value is strongly dependend on the seed (~5-10% deviation)
    # TODO: propagate precision to kedro parameters
    if len(circuit.parameters) == 0:
        expressibility = 0
    else:
        hi = get_haar_integral(
            n_qubits=circuit.num_qubits, samples=haar_samples_per_qubit, rng=jrng
        )
        pi = pqc_integral(
            circuit=circuit,
            samples=samples_per_parameter * int(np.log(len(circuit.parameters)) + 1)
            + haar_samples_per_qubit * circuit.num_qubits,
            params_shape=len(circuit.parameters),
            precision=5,
            rng=rng,
        )
        expressibility = jf(hi, pi)

    return {
        "expressibility": expressibility,
    }


def calculate_measures(
    circuit: QuantumCircuit,
    samples_per_parameter: int,
    haar_samples_per_qubit: int,
    seed: int,
) -> List[float]:
    expressibility = calculate_expressibility(
        circuit=circuit,
        samples_per_parameter=samples_per_parameter,
        haar_samples_per_qubit=haar_samples_per_qubit,
        seed=seed,
    )["expressibility"]

    entangling_capability = calculate_entangling_capability(
        circuit=circuit,
        samples_per_parameter=samples_per_parameter,
        samples_per_qubit=haar_samples_per_qubit,
        seed=seed,
    )["entangling_capability"]

    return {
        "measure": pd.DataFrame(
            {
                "expressibility": [expressibility],
                "entangling_capability": [entangling_capability],
            }
        )
    }



import time
import re
import numpy as np

import dask.array as da

import pennylane as qml

import qiskit
from qiskit.quantum_info import Operator

import qulacs
from qulacs import QuantumCircuit as QualcsQuantumCircuit
from qulacs.gate import DenseMatrix, Identity

import qibo

import qrisp

import cirq
from cirq.contrib.qasm_import import circuit_from_qasm


from typing import Dict, List


def calculate_n_qubits_from_qasm(qasm_string):
    return int(
        re.findall(r"qreg q\[(?P<n_qubits>\d*)\]", qasm_string)[0]
    )  # TODO: improvement wanted


class test_fw:
    time_const = 1e-9
    constant_sleep = True
    load = True

    def __init__(self, qasm_circuit, n_shots):
        self.n_qubits = calculate_n_qubits_from_qasm(qasm_circuit)

        self.depth = int(
            qasm_circuit[qasm_circuit.find("\nqreg q[") + 8]
        )  # TODO: improvement wanted

        self.shots = n_shots

    def execute(self) -> None:
        if self.constant_sleep:
            if self.load:
                # Following https://tutorial.dask.org/02_array.html#Dask-array-version
                xd = da.random.normal(10, 0.1, size=(300, 300), chunks=(30, 30))
                yd = xd.mean(axis=0)
                yd.compute()
            else:
                time.sleep(0.01)
        else:
            time.sleep(
                self.time_const * self.shots * self.depth**2 * self.n_qubits**3
            )

    def get_result(self) -> Dict[str, float]:
        counts = {}

        for i in range(2**self.n_qubits):
            bitstring = format(i, f"0{self.n_qubits}b")
            counts[bitstring] = 0

        return counts


class pennylane_fw:
    def __init__(self, qasm_circuit, n_shots):
        self.n_qubits = calculate_n_qubits_from_qasm(qasm_circuit)

        self.n_shots = n_shots

        if self.n_qubits <= 20:
            self.backend = qml.device(
                "default.qubit",
                wires=range(self.n_qubits),
                shots=self.n_shots,
                max_workers=None,  # restrict subworkers to 1 to prevent issues with MP https://docs.pennylane.ai/en/stable/code/api/pennylane.devices.default_qubit.DefaultQubit.html # noqa
            )
        else:  # recommended to be used for > 20 qubits
            self.backend = qml.device(
                "lightning.qubit", wires=range(self.n_qubits), shots=self.n_shots
            )

        # Pennylane does not support measurements at the moment
        qasm_circuit_wo_measurement = re.sub(r"measure.*;\n", "", qasm_circuit)
        self.qml_qasm = qml.from_qasm(qasm_circuit_wo_measurement)

        @qml.qnode(self.backend)
        def circuit():
            self.qml_qasm()
            return qml.counts()

        self.qc = circuit

    def execute(self) -> None:
        self.result = self.qc()

    def get_result(self) -> Dict[str, float]:
        counts = self.result

        for i in range(2**self.n_qubits):
            bitstring = format(i, f"0{self.n_qubits}b")
            if bitstring not in counts.keys():
                counts[bitstring] = 0

            else:
                counts[bitstring] = counts[bitstring] / self.n_shots

        return counts


class qiskit_fw:
    def __init__(self, qasm_circuit, n_shots):
        # self.backend = qiskit.Aer.get_backend("qasm_simulator")
        self.backend = qiskit.Aer.get_backend("aer_simulator")

        self.n_qubits = calculate_n_qubits_from_qasm(qasm_circuit)

        self.qc = qiskit.QuantumCircuit.from_qasm_str(qasm_circuit)
        self.n_shots = n_shots
        self.result = None

    def execute(self) -> None:
        self.result = qiskit.execute(self.qc, backend=self.backend, shots=self.n_shots)

    def get_result(self) -> Dict[str, float]:
        counts = self.result.result().get_counts()

        for i in range(2**self.n_qubits):
            bitstring = format(i, f"0{self.n_qubits}b")
            if bitstring not in counts.keys():
                counts[bitstring] = 0

        return counts


class qrisp_fw:
    def __init__(self, qasm_circuit, n_shots):
        self.n_qubits = calculate_n_qubits_from_qasm(qasm_circuit)

        self.qc = qrisp.QuantumCircuit.from_qasm_str(qasm_circuit)
        self.n_shots = n_shots
        self.result = None

    def execute(self) -> None:
        self.result = self.qc.run(shots=self.n_shots)

    def get_result(self) -> Dict[str, float]:
        counts = self.result

        for i in range(2**self.n_qubits):
            bitstring = format(i, f"0{self.n_qubits}b")
            if bitstring not in counts.keys():
                counts[bitstring] = 0

        return counts


class numpy_fw:
    def __init__(self, qasm_circuit, n_shots):
        self.n_shots = n_shots
        cicruit = qiskit.QuantumCircuit.from_qasm_str(qasm_circuit)
        cicruit.remove_final_measurements()
        self.n_qubits = calculate_n_qubits_from_qasm(qasm_circuit)
        self.qc = Operator(cicruit)

    def execute(self) -> None:
        # keine circuits sondern fertige Matrizen
        statevector = np.array(self.qc)[:, 0]
        probabilities = np.abs((statevector) ** 2)
        if self.n_shots is not None:
            self.result = np.random.choice(
                len(probabilities), self.n_shots, p=probabilities
            )
        else:
            self.result = probabilities

    def get_result(self) -> Dict[str, float]:
        counts = {}

        # TODO verification needed!
        for i in range(2**self.n_qubits):
            bitstring = format(i, f"0{self.n_qubits}b")
            c = np.count_nonzero(self.result == i)
            counts[bitstring] = c

        return counts


class cirq_fw:
    def __init__(self, qasm_circuit, n_shots):
        self.n_shots = n_shots

        self.n_qubits = calculate_n_qubits_from_qasm(qasm_circuit)

        self.qc = circuit_from_qasm(qasm_circuit)

        self.qc.append(
            cirq.measure(cirq.NamedQubit.range(self.n_qubits, prefix=""), key="result")
        )
        self.backend = cirq.Simulator()

    def execute(self) -> None:
        if self.n_shots is None:
            self.result = self.backend.simulate(self.qc)
        else:
            self.result = self.backend.run(self.qc, repetitions=self.n_shots)

    def get_result(self) -> Dict[str, float]:
        counts = {}

        # TODO verification needed!
        for i in range(2**self.n_qubits):
            bitstring = format(i, f"0{self.n_qubits}b")

            mask = None
            for j in range(self.n_qubits):
                if mask is None:
                    mask = self.result.data[f"c_{j}"] == int(bitstring[j])
                else:
                    mask &= self.result.data[f"c_{j}"] == int(bitstring[j])
            result = len(self.result.data[mask])

            counts[bitstring] = result / self.n_shots

        return counts


class qibo_fw:
    def __init__(self, qasm_circuit, n_shots):
        qibo.set_backend("numpy")
        self.backend = qibo.get_backend()
        self.n_shots = n_shots

        self.n_qubits = calculate_n_qubits_from_qasm(qasm_circuit)

        # this is super hacky, but the way qibo parses the QASM string
        # does not deserve better.
        def qasm_conv(match: re.Match) -> str:
            denominator = float(match.group()[1:])
            return f"*{1/denominator}"

        qasm_circuit = re.sub(r"/\d*", qasm_conv, qasm_circuit, flags=re.MULTILINE)

        self.qc = qibo.models.Circuit.from_qasm(qasm_circuit)

    def execute(self) -> None:
        self.result = self.qc(nshots=self.n_shots)

    def get_result(self) -> Dict[str, float]:
        counts = dict(self.result.frequencies(binary=True))

        # TODO verification needed!
        for i in range(2**self.n_qubits):
            bitstring = format(i, f"0{self.n_qubits}b")
            if bitstring not in counts.keys():
                counts[bitstring] = 0

        return counts


class qulacs_fw:
    def __init__(self, qasm_circuit, n_shots):
        self.n_qubits = calculate_n_qubits_from_qasm(qasm_circuit)
        # removing measurment operation...
        qasm_circuit = re.sub("\\nmeasure .*;", "", qasm_circuit)
        # ... and creg operation as they are not supported by qulacs
        # see `convert_QASM_to_qulacs_circuit` for details.
        qasm_circuit = re.sub("\\ncreg .*;", "", qasm_circuit)
        self.qc = self.convert_QASM_to_qulacs_circuit(qasm_circuit.split("\n"))
        self.n_shots = n_shots
        self.result = None

    def execute(self) -> None:
        state = qulacs.QuantumState(self.n_qubits)
        self.qc.update_quantum_state(state)
        self.result = state.sampling(sampling_count=self.n_shots)

    def get_result(self) -> Dict[str, float]:
        counts = {}
        for i in range(2**self.n_qubits):
            bitstring = format(i, f"0{self.n_qubits}b")
            if i not in self.result:
                counts[bitstring] = 0
            else:
                counts[bitstring] = sum(c == i for c in self.result)

        return counts

    def convert_QASM_to_qulacs_circuit(
        self, input_strs: List[str], *, remap_remove: bool = False
    ) -> QualcsQuantumCircuit:
        """
        Copied from https://github.com/qulacs/qulacs/blob/2750414e8e1ebc61064207158dd8b0e618708ad4/pysrc/qulacs/converter/qasm_converter.py#L120 # noqa
        Only difference is the missing bracket for the u2 parsing

        convert QASM List[str] to qulacs QuantumCircuit.

        constraints: qreg must be named q, and creg cannot be used.
        """
        FIXED_POINT_PATTERN = r"[+-]?\d+(?:\.\d*)?|\.\d+"
        FLOATING_POINT_PATTERN = r"[eE][-+]?\d+"
        GENERAL_NUMBER_PATTERN = (
            rf"(?:{FIXED_POINT_PATTERN})(?:{FLOATING_POINT_PATTERN})?"  # noqa
        )

        mapping: List[int] = []

        for instr_moto in input_strs:
            # process input string for parsing instruction.
            instr = instr_moto.lower().strip().replace(" ", "").replace("\t", "")
            if instr == "":
                continue
            if instr[0:4] == "qreg":
                matchobj = re.match(r"qregq\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir = QualcsQuantumCircuit(int(ary[0]))
                if len(mapping) == 0:
                    mapping = list(range(int(ary[0])))
            elif instr[0:2] == "cx":
                matchobj = re.match(r"cxq\[(\d+)\],q\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_CNOT_gate(mapping[int(ary[0])], mapping[int(ary[1])])
            elif instr[0:2] == "cz":
                matchobj = re.match(r"czq\[(\d+)\],q\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_CZ_gate(mapping[int(ary[0])], mapping[int(ary[1])])
            elif instr[0:4] == "swap":
                matchobj = re.match(r"swapq\[(\d+)\],q\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_SWAP_gate(mapping[int(ary[0])], mapping[int(ary[1])])
            elif instr[0:2] == "id":
                matchobj = re.match(r"idq\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_gate(Identity(mapping[int(ary[0])]))
            elif instr[0:2] == "xq":
                matchobj = re.match(r"xq\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_X_gate(mapping[int(ary[0])])
            elif instr[0:2] == "yq":
                matchobj = re.match(r"yq\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_Y_gate(mapping[int(ary[0])])
            elif instr[0:2] == "zq":
                matchobj = re.match(r"zq\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_Z_gate(mapping[int(ary[0])])
            elif instr[0:2] == "hq":
                matchobj = re.match(r"hq\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_H_gate(mapping[int(ary[0])])
            elif instr[0:2] == "sq":
                matchobj = re.match(r"sq\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_S_gate(mapping[int(ary[0])])
            elif instr[0:4] == "sdgq":
                matchobj = re.match(r"sdgq\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_Sdag_gate(mapping[int(ary[0])])
            elif instr[0:2] == "tq":
                matchobj = re.match(r"tq\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_T_gate(mapping[int(ary[0])])
            elif instr[0:4] == "tdgq":
                matchobj = re.match(r"tdgq\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_Tdag_gate(mapping[int(ary[0])])
            elif instr[0:2] == "rx":
                matchobj = re.match(
                    rf"rx\(({GENERAL_NUMBER_PATTERN})\)q\[(\d+)\];", instr
                )
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_RX_gate(mapping[int(ary[1])], -float(ary[0]))
            elif instr[0:2] == "ry":
                matchobj = re.match(
                    rf"ry\(({GENERAL_NUMBER_PATTERN})\)q\[(\d+)\];", instr
                )
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_RY_gate(mapping[int(ary[1])], -float(ary[0]))
            elif instr[0:2] == "rz":
                matchobj = re.match(
                    rf"rz\(({GENERAL_NUMBER_PATTERN})\)q\[(\d+)\];", instr
                )
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_RZ_gate(mapping[int(ary[1])], -float(ary[0]))
            elif instr[0:1] == "p":
                matchobj = re.match(rf"p\({GENERAL_NUMBER_PATTERN}\)q\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_U1_gate(mapping[int(ary[1])], float(ary[0]))
            elif instr[0:2] == "u1":
                matchobj = re.match(
                    rf"u1\(({GENERAL_NUMBER_PATTERN})\)q[(\d+)];", instr
                )
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_U1_gate(mapping[int(ary[1])], float(ary[0]))
            elif instr[0:2] == "u2":
                matchobj = re.match(
                    rf"u2\(({GENERAL_NUMBER_PATTERN}),"
                    + rf"({GENERAL_NUMBER_PATTERN})\)q\[(\d+)\];",
                    instr,
                )
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_U2_gate(mapping[int(ary[2])], float(ary[0]), float(ary[1]))
            elif instr[0:2] == "u3":
                matchobj = re.match(
                    rf"u3\(({GENERAL_NUMBER_PATTERN}),"
                    + rf"({GENERAL_NUMBER_PATTERN}),"
                    + rf"({GENERAL_NUMBER_PATTERN})\)q\[(\d+)\];",
                    instr,
                )
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_U3_gate(
                    mapping[int(ary[3])], float(ary[0]), float(ary[1]), float(ary[2])
                )
            elif instr[0:1] == "u":
                matchobj = re.match(
                    rf"u\(({GENERAL_NUMBER_PATTERN}),"
                    + rf"({GENERAL_NUMBER_PATTERN}),"
                    + rf"({GENERAL_NUMBER_PATTERN})\)q\[(\d+)\];",
                    instr,
                )
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_U3_gate(
                    mapping[int(ary[3])], float(ary[0]), float(ary[1]), float(ary[2])
                )
            elif instr[0:3] == "sxq":
                matchobj = re.match(r"sxq\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_sqrtX_gate(mapping[int(ary[0])])
            elif instr[0:5] == "sxdgq":
                matchobj = re.match(r"sxdgq\[(\d+)\];", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                cir.add_sqrtXdag_gate(mapping[int(ary[0])])
            elif instr[0:11] == "densematrix":
                # Matches all matrix elements and qubit indexes
                deary = re.findall(r"[+-]?\d+(?:\.\d+)?(?:[eE][+-]?\d+)?", instr)
                target_qubit_count = int(deary[0])
                control_qubit_count = int(deary[1])

                gate_mat = np.zeros(
                    (2**target_qubit_count, 2**target_qubit_count), dtype="complex"
                )
                bas = 2
                for i in range(2**target_qubit_count):
                    for j in range(2**target_qubit_count):
                        gate_mat[i][j] = (
                            float(deary[bas]) + float(deary[bas + 1]) * 1.0j
                        )
                        bas += 2
                control_values = []
                for i in range(control_qubit_count):
                    control_values.append(mapping[int(deary[bas])])
                    bas += 1
                terget_indexes = []
                for i in range(target_qubit_count):
                    terget_indexes.append(mapping[int(deary[bas])])
                    bas += 1

                dense_gate = DenseMatrix(terget_indexes, gate_mat)  # type:ignore
                for i in range(control_qubit_count):
                    control_index = int(deary[bas])
                    bas += 1
                    dense_gate.add_control_qubit(control_index, control_values[i])
                cir.add_gate(dense_gate)
            elif remap_remove and instr[0:4] == "//q[":
                matchobj = re.match(r"//q\[(\d+)-->q\[(\d+)\]", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                if not (ary is None):
                    mapping[int(ary[0])] = int(ary[1])
            elif remap_remove and instr[0:8] == "//qubits":
                matchobj = re.match(r"//qubits:(\d+)", instr)
                assert matchobj is not None
                ary = matchobj.groups()
                mapping = list(range(int(ary[0])))
            elif instr == "openqasm2.0;" or instr == 'include"qelib1.inc";':
                # related to file format, not for read.
                pass
            else:
                raise RuntimeError(f"unknown line: {instr}")
        return cir



"""
This is a boilerplate pipeline 'data_science'
generated using Kedro 0.18.3
"""

from kedro.pipeline import node, pipeline
from quafel.pipelines.data_science.nodes import (
    measure_execution_durations,
    combine_evaluations,
)


def create_pipeline(partitions, **kwargs) -> dict:
    pl_parallel_measure_execution_durations = pipeline(
        [
            *[
                node(
                    func=measure_execution_durations,
                    inputs={
                        "evaluations": "params:evaluations",
                        f"qasm_circuit_{i}": f"qasm_circuit_{i}",
                        f"n_shots_{i}": f"n_shots_{i}",
                        f"framework_id_{i}": f"framework_{i}",
                    },
                    outputs={
                        "execution_duration": f"execution_duration_{i}",
                        "execution_result": f"execution_result_{i}",
                    },
                    tags=["dynamic"],
                    name=f"measure_execution_durations_{i}",
                )
                for i in partitions
            ],
        ],
        inputs={
            **{f"qasm_circuit_{i}": f"qasm_circuit_{i}" for i in partitions},
            **{f"n_shots_{i}": f"n_shots_{i}" for i in partitions},
            **{f"framework_{i}": f"framework_{i}" for i in partitions},
        },
        outputs={
            **{
                f"execution_duration_{i}": f"execution_duration_{i}" for i in partitions
            },
            **{f"execution_result_{i}": f"execution_result_{i}" for i in partitions},
        },
        namespace="data_science",
    )

    pl_combine_evaluations = pipeline(
        [
            node(
                func=combine_evaluations,
                inputs={
                    "evaluation_partitions": "evaluation_partitions",
                    "execution_durations": "execution_durations",
                    "execution_results": "execution_results",
                    "measures": "measures",
                    "export_results": "params:export_results",
                },
                outputs={
                    "evaluations_combined": "evaluations_combined",
                },
                tags=["static"],
                name="combine_evaluations",
            ),
        ],
        inputs={
            "evaluation_partitions": "evaluation_partitions",
            "execution_durations": "execution_durations",
            "execution_results": "execution_results",
            "measures": "measures",
        },
        outputs={
            "evaluations_combined": "evaluations_combined",
        },
        namespace="data_science",
    )

    return {
        "pl_parallel_measure_execution_durations": pl_parallel_measure_execution_durations,  # noqa
        "pl_combine_evaluations": pl_combine_evaluations,
    }



"""Command line tools for manipulating a Kedro project.
Intended to be invoked via `kedro`."""
import click
from kedro.framework.cli.project import (
    ASYNC_ARG_HELP,
    CONFIG_FILE_HELP,
    CONF_SOURCE_HELP,
    FROM_INPUTS_HELP,
    FROM_NODES_HELP,
    LOAD_VERSION_HELP,
    NODE_ARG_HELP,
    PARAMS_ARG_HELP,
    PIPELINE_ARG_HELP,
    RUNNER_ARG_HELP,
    TAG_ARG_HELP,
    TO_NODES_HELP,
    TO_OUTPUTS_HELP,
    project_group,
)
from kedro.framework.cli.utils import (
    CONTEXT_SETTINGS,
    _config_file_callback,
    _get_values_as_tuple,
    _reformat_load_versions,
    _split_params,
    env_option,
    split_string,
    split_node_names,
)
from kedro.framework.session import KedroSession
from kedro.utils import load_obj


@click.group(context_settings=CONTEXT_SETTINGS, name=__file__)
def cli():
    """Command line tools for manipulating a Kedro project."""


@project_group.command()
@click.option(
    "--from-inputs", type=str, default="", help=FROM_INPUTS_HELP, callback=split_string
)
@click.option(
    "--to-outputs", type=str, default="", help=TO_OUTPUTS_HELP, callback=split_string
)
@click.option(
    "--from-nodes",
    type=str,
    default="",
    help=FROM_NODES_HELP,
    callback=split_node_names,
)
@click.option(
    "--to-nodes", type=str, default="", help=TO_NODES_HELP, callback=split_node_names
)
@click.option("--node", "-n", "node_names", type=str, multiple=True, help=NODE_ARG_HELP)
@click.option(
    "--runner", "-r", type=str, default=None, multiple=False, help=RUNNER_ARG_HELP
)
@click.option("--async", "is_async", is_flag=True, multiple=False, help=ASYNC_ARG_HELP)
@env_option
@click.option("--tag", "-t", type=str, multiple=True, help=TAG_ARG_HELP)
@click.option(
    "--load-version",
    "-lv",
    type=str,
    multiple=True,
    help=LOAD_VERSION_HELP,
    callback=_reformat_load_versions,
)
@click.option("--pipeline", "-p", type=str, default=None, help=PIPELINE_ARG_HELP)
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help=CONFIG_FILE_HELP,
    callback=_config_file_callback,
)
@click.option(
    "--conf-source",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help=CONF_SOURCE_HELP,
)
@click.option(
    "--params",
    type=click.UNPROCESSED,
    default="",
    help=PARAMS_ARG_HELP,
    callback=_split_params,
)
# pylint: disable=too-many-arguments,unused-argument
def run(
    tag,
    env,
    runner,
    is_async,
    node_names,
    to_nodes,
    from_nodes,
    from_inputs,
    to_outputs,
    load_version,
    pipeline,
    config,
    conf_source,
    params,
):
    """Run the pipeline."""

    runner = runner or "SequentialRunner"

    tag = _get_values_as_tuple(tag) if tag else tag
    node_names = _get_values_as_tuple(node_names) if node_names else node_names

    def _instantiate_runner(runner, is_async, project_context):
        runner_class = load_obj(runner, "kedro.runner")
        runner_kwargs = dict(is_async=is_async)

        if runner.endswith("DaskRunner"):
            client_args = project_context.params.get("dask_client") or {}
            runner_kwargs.update(client_args=client_args)

        return runner_class(**runner_kwargs)

    with KedroSession.create(env=env, extra_params=params) as session:
        context = session.load_context()
        runner_instance = _instantiate_runner(runner, is_async, context)
        session.run(
            tags=tag,
            runner=runner_instance,
            node_names=node_names,
            from_nodes=from_nodes,
            to_nodes=to_nodes,
            from_inputs=from_inputs,
            to_outputs=to_outputs,
            load_versions=load_version,
            pipeline_name=pipeline,
        )



"""Project pipelines."""

from typing import Dict

from kedro.pipeline import Pipeline

from quafel.pipelines import data_generation as dg
from quafel.pipelines import data_science as ds
from quafel.pipelines import visualization as viz

import glob
from pathlib import Path


def register_pipelines() -> Dict[str, Pipeline]:
    """Register the project's pipelines.

    Returns:
        A mapping from pipeline names to ``Pipeline`` objects.
    """
    # get all the partitions created with the "prepare" pipeline
    all_partitions = [Path(f).stem for f in glob.glob("data/02_intermediate/*.csv")]

    # ----------------------------------------------------------------
    # Gather all the existing artifacts and calculate the difference
    # ----------------------------------------------------------------

    # get all existing circuits, durations and results.
    # The hooks run prior to this, so
    # in case we don't want to restore existing results,
    # we should find empty directories
    existing_circuits = [Path(f).stem for f in glob.glob("data/03_qasm_circuits/*.txt")]
    # get all existing durations and results. The hooks run prior to this,
    # so in case we don't want to restore existing results,
    # we should find empty directories
    existing_measures = [Path(f).stem for f in glob.glob("data/04_measures/*.csv")]
    existing_durations = [
        Path(f).stem for f in glob.glob("data/06_execution_durations/*.csv")
    ]
    existing_results = [
        Path(f).stem for f in glob.glob("data/05_execution_results/*.csv")
    ]

    # get the intersection of the durations and results
    existing_evals = [m for m in existing_durations if m in existing_results]
    # .. and the intersection of all partitions
    ds_partitions = [p for p in all_partitions if p not in existing_evals]

    # gather all the .tmp files to create figures output
    tmp_files = [Path(f).stem for f in glob.glob("data/08_reporting/*.tmp")]

    # ----------------------------------------------------------------
    # Create the pipelines
    # ----------------------------------------------------------------

    # pass only the number of partitions we want to generate circuits for
    dg_pipelines = dg.create_pipeline(
        partitions=all_partitions,
        existing_circuits=existing_circuits,
        existing_measures=existing_measures,
    )
    # pass only the number of partitions we want to evaluate
    # (this would be equal to all partitions in an initial run or
    # in case we don't want to restore existing results)
    ds_pipelines = ds.create_pipeline(partitions=ds_partitions)
    viz_pipelines = viz.create_pipeline(figures=tmp_files)

    return {
        "__default__": dg_pipelines["pl_generate_evaluation_partitions"]
        + dg_pipelines["pl_generate_qasm_circuits"]
        + ds_pipelines["pl_parallel_measure_execution_durations"]
        + viz_pipelines["pl_visualize_evaluations"],
        "prepare": dg_pipelines["pl_generate_evaluation_partitions"],
        "measure": dg_pipelines[
            # "pl_generate_qasm_circuits"
            "pl_generate_qasm_circuits"
        ]
        + ds_pipelines["pl_parallel_measure_execution_durations"],
        # ct measure is the same as the measure pipeline, but we need to
        # tell the hooks that we don't want to delete the existing results
        "combine": ds_pipelines["pl_combine_evaluations"],
        "visualize": viz_pipelines["pl_visualize_evaluations"],
        "print_tests": viz_pipelines["pl_print_tests"],
    }



"""
This is a boilerplate pipeline 'visualization'
generated using Kedro 0.18.4
"""

from kedro.pipeline import node, pipeline
from quafel.pipelines.visualization.nodes import (
    shots_depths_viz,
    shots_qubits_viz,
    depth_qubits_viz,
    qubits_time_viz,
    depth_time_viz,
    shots_time_viz,
    depth_measures_viz,
    qubits_measures_viz,
    shots_measures_viz,
    export_selected,
    extract_tests,
)


def create_pipeline(figures, **kwargs) -> dict:
    pl_visualize_evaluations = pipeline(
        [
            node(
                func=shots_depths_viz,
                inputs={
                    "evaluations_combined": "evaluations_combined",
                },
                outputs={
                    **{
                        f: f
                        for f in filter(
                            lambda s: ("_fw" in s)
                            and ("depth_" in s)
                            and ("time" in s),
                            figures,
                        )
                    },
                },
                tags=["static"],
                name="shots_depths_viz",
            ),
            node(
                func=shots_qubits_viz,
                inputs={
                    "evaluations_combined": "evaluations_combined",
                },
                outputs={
                    **{
                        f: f
                        for f in filter(
                            lambda s: ("_fw" in s)
                            and ("qubits_" in s)
                            and ("time" in s),
                            figures,
                        )
                    },
                },
                tags=["static"],
                name="shots_depth_viz",
            ),
            node(
                func=depth_qubits_viz,
                inputs={
                    "evaluations_combined": "evaluations_combined",
                },
                outputs={
                    **{
                        f: f
                        for f in filter(
                            lambda s: ("_fw" in s)
                            and ("shots_" in s)
                            and ("time" in s),
                            figures,
                        )
                    },
                },
                tags=["static"],
                name="depth_qubits_viz",
            ),
            node(
                func=qubits_time_viz,
                inputs={
                    "evaluations_combined": "evaluations_combined",
                    "skip_frameworks": "params:skip_frameworks",
                },
                outputs={
                    **{
                        f: f
                        for f in filter(
                            lambda s: ("shots_" in s)
                            and ("depth_" in s)
                            and ("time" in s),
                            figures,
                        )
                    },
                },
                tags=["static"],
                name="qubits_time_viz",
            ),
            node(
                func=shots_time_viz,
                inputs={
                    "evaluations_combined": "evaluations_combined",
                    "skip_frameworks": "params:skip_frameworks",
                },
                outputs={
                    **{
                        f: f
                        for f in filter(
                            lambda s: ("qubits_" in s)
                            and ("depth_" in s)
                            and ("time" in s),
                            figures,
                        )
                    },
                },
                tags=["static"],
                name="shots_time_viz",
            ),
            node(
                func=depth_time_viz,
                inputs={
                    "evaluations_combined": "evaluations_combined",
                    "skip_frameworks": "params:skip_frameworks",
                },
                outputs={
                    **{
                        f: f
                        for f in filter(
                            lambda s: ("shots_" in s)
                            and ("qubits_" in s)
                            and ("time" in s),
                            figures,
                        )
                    },
                },
                tags=["static"],
                name="depth_time_viz",
            ),
            node(
                func=depth_measures_viz,
                inputs={
                    "evaluations_combined": "evaluations_combined",
                },
                outputs={
                    **{
                        f: f
                        for f in filter(
                            lambda s: ("shots_" in s)
                            and ("qubits_" in s)
                            and ("measures" in s),
                            figures,
                        )
                    },
                },
                tags=["static"],
                name="depth_measures_viz",
            ),
            node(
                func=shots_measures_viz,
                inputs={
                    "evaluations_combined": "evaluations_combined",
                },
                outputs={
                    **{
                        f: f
                        for f in filter(
                            lambda s: ("qubits_" in s)
                            and ("depth_" in s)
                            and ("measures" in s),
                            figures,
                        )
                    },
                },
                tags=["static"],
                name="shots_measures_viz",
            ),
            node(
                func=qubits_measures_viz,
                inputs={
                    "evaluations_combined": "evaluations_combined",
                },
                outputs={
                    **{
                        f: f
                        for f in filter(
                            lambda s: ("shots_" in s)
                            and ("depth_" in s)
                            and ("measures" in s),
                            figures,
                        )
                    },
                },
                tags=["static"],
                name="qubits_measures_viz",
            ),
        ],
        inputs={
            "evaluations_combined": "evaluations_combined",
        },
        outputs={
            **{f: f for f in figures},
        },
        namespace="visualization",
    )

    pl_print_tests = pipeline(
        [
            node(
                func=extract_tests,
                inputs={
                    "evaluations_combined": "evaluations_combined",
                },
                outputs={},
                tags=["static"],
                name="extract_tests",
            )
        ],
        inputs={
            "evaluations_combined": "evaluations_combined",
        },
        outputs={},
        namespace="visualization",
    )

    pl_export_visualizations = pipeline(
        [
            node(
                func=export_selected,
                inputs={
                    "evaluations_combined": "evaluations_combined",
                    "additional_figures": "params:additional_figures",
                    "output_folder": "params:output_folder",
                    **{f: f for f in figures},
                },
                outputs={},
                tags=["dynamic"],
                name="export_selected",
            )
        ],
        inputs={
            "evaluations_combined": "evaluations_combined",
            **{f: f for f in figures},
        },
        namespace="visualization",
    )

    return {
        "pl_visualize_evaluations": pl_visualize_evaluations + pl_export_visualizations,
        "pl_print_tests": pl_print_tests,
    }



from kedro.extras.datasets.pandas import CSVDataSet
from kedro.extras.datasets.text import TextDataSet
from kedro.extras.datasets.plotly import JSONDataSet
from kedro.io import Version
from kedro.framework.hooks import hook_impl

from typing import Any, Dict

import glob
from pathlib import Path

import os
import logging
from kedro.io import DataCatalog

import time

log = logging.getLogger(__name__)


class ProjectHooks:
    @hook_impl
    def after_context_created(self, context):
        pass


class PipelineHooks:
    @property
    def _logger(self):
        return logging.getLogger(self.__class__.__name__)

    @hook_impl
    def before_pipeline_run(self, run_params: Dict[str, Any], pipeline, catalog):
        """A hook implementation to add a catalog entry
        based on the filename passed to the command line, e.g.:
            kedro run --params=input:iris_1.csv
            kedro run --params=input:iris_2.csv
            kedro run --params=input:iris_3.csv
        """

        # ----------------------------------------------------------------
        # If we are running a new experiment setup (e.g. executing the "pre"
        # pipeline), we want to cleanup all the files that are **not**
        # versioned but were used as results in previous runs
        # ----------------------------------------------------------------

        if run_params["pipeline_name"] == "prepare":
            tempFiles = glob.glob("data/02_intermediate/*.csv")
            for f in tempFiles:
                os.remove(f)
            # tempFiles = glob.glob("data/08_reporting/*.tmp")
            # for f in tempFiles:
            #     os.remove(f)

        # cleanup "results" if we are at the beginning our our experiment
        if (
            run_params["pipeline_name"] is None  # Running the Default pipeline
            or run_params["pipeline_name"] == "prepare"
        ):
            tempFiles = glob.glob("data/03_qasm_circuits/*.txt")
            for f in tempFiles:
                os.remove(f)
            tempFiles = glob.glob("data/04_measures/*.csv")
            for f in tempFiles:
                os.remove(f)
            tempFiles = glob.glob("data/05_execution_results/*.csv")
            for f in tempFiles:
                os.remove(f)

            tempFiles = glob.glob("data/06_execution_durations/*.csv")
            for f in tempFiles:
                os.remove(f)

        self.start_run = time.time()

    @hook_impl
    def after_pipeline_run(self, run_params: Dict[str, Any], pipeline, catalog):
        self.finish_run = time.time()

        log.info(f"Run took {self.finish_run - self.start_run}s")

        # after visualization, we don't need the tmp files anymore
        # if (
        #     run_params["pipeline_name"] is None  # Running the Default pipeline
        #     or run_params["pipeline_name"] == "visualize"
        # ):

        # only cleanup if the last (visualize) pipeline ran
        if (
            run_params["pipeline_name"] is None
            or run_params["pipeline_name"] == "visualize"
        ):
            tempFiles = glob.glob("data/02_intermediate/*.csv")
            for f in tempFiles:
                os.remove(f)
            tempFiles = glob.glob("data/03_qasm_circuits/*.txt")
            for f in tempFiles:
                os.remove(f)
            tempFiles = glob.glob("data/04_measures/*.csv")
            for f in tempFiles:
                os.remove(f)
            tempFiles = glob.glob("data/05_execution_results/*.csv")
            for f in tempFiles:
                os.remove(f)
            tempFiles = glob.glob("data/06_execution_durations/*.csv")
            for f in tempFiles:
                os.remove(f)
            tempFiles = glob.glob("data/08_reporting/*.tmp")
            for f in tempFiles:
                os.remove(f)

        if run_params["pipeline_name"] == "prepare":
            # Cleanup all previously generated temp files,
            # we will generate them again below
            tempFiles = glob.glob("data/08_reporting/*.tmp")
            for f in tempFiles:
                os.remove(f)
            # ----------------------------------------------------------------
            # This section ensures that the reporting dictionary always
            # contains the proper output data catalogs so that kedro-viz is happy
            # ----------------------------------------------------------------

            try:
                # get the evaluation matrix where information about
                # all the possible combinations is stored
                evaluation_matrix = (
                    catalog.datasets.data_generation__evaluation_matrix.load()
                )
            except Exception as e:
                log.exception(e)
                return

            # generate a list of names that will be used as plots
            # later in the visualization pipeline
            # If you add new visualization outputs, you must also
            # create the file names here
            names = []
            for fw in evaluation_matrix["frameworks"]:
                for q in evaluation_matrix["qubits"]:
                    names.append(f"{fw}_qubits_{q}_time")
                for d in evaluation_matrix["depths"]:
                    names.append(f"{fw}_depth_{d}_time")
                for s in evaluation_matrix["shots"]:
                    names.append(f"{fw}_shots_{s}_time")

            for s in evaluation_matrix["shots"]:
                for d in evaluation_matrix["depths"]:
                    names.append(f"shots_{s}_depth_{d}_time")
                    names.append(f"shots_{s}_depth_{d}_measures")

            for d in evaluation_matrix["depths"]:
                for q in evaluation_matrix["qubits"]:
                    names.append(f"qubits_{q}_depth_{d}_time")
                    names.append(f"qubits_{q}_depth_{d}_measures")

            for q in evaluation_matrix["qubits"]:
                for s in evaluation_matrix["shots"]:
                    names.append(f"shots_{s}_qubits_{q}_time")
                    names.append(f"shots_{s}_qubits_{q}_measures")

            # use the dummy dataset to get the version of the current
            # kedro run, so that it matches the ones from the versioned datasets
            version = Version(
                None, catalog.datasets.dummy_versioned_dataset._version.save
            )

            # iterate all the names and create a JSONDataSet (plotly) for each one
            for name in names:
                filepath = os.path.join("data/08_reporting/", f"{name}.json")

                dataset_template = JSONDataSet(filepath=filepath, version=version)
                catalog.add(name, dataset_template)

                # create a dictionary if necessary
                # (versioned datasets need dictionaries)
                try:
                    os.mkdir(filepath)
                except FileExistsError:
                    # directory already exists
                    pass

                # create a .tmp file which we will use later in the
                # pipeline_registry to create node outputs dynamically
                with open(os.path.join("data/08_reporting/", f"{name}.tmp"), "w") as fw:
                    fw.write("")

    @hook_impl
    def on_pipeline_error(self, run_params: Dict[str, Any], pipeline, catalog):
        # tempFiles = glob.glob("data/02_intermediate/*.csv")
        # for f in tempFiles:
        #     os.remove(f)
        # tempFiles = glob.glob("data/05_execution_results/*.csv")
        # for f in tempFiles:
        #     os.remove(f)
        # tempFiles = glob.glob("data/06_execution_durations/*.csv")
        # for f in tempFiles:
        #     os.remove(f)
        tempFiles = glob.glob("data/08_reporting/*.tmp")
        for f in tempFiles:
            os.remove(f)


class DataCatalogHooks:
    @property
    def _logger(self):
        return logging.getLogger(self.__class__.__name__)

    @hook_impl
    def after_catalog_created(self, catalog: DataCatalog) -> None:
        # ----------------------------------------------------------------
        # This section creates csv datasets based on the partitioned
        # dataset stored in the intermediate directory
        # Also based on those partitions, it creates further csv datasets
        # for the evaluation durations and the evaluation results
        # ----------------------------------------------------------------

        # get all the partitions created by the "pre" pipeline
        partitions = glob.glob("data/02_intermediate/*.csv")

        for partition in partitions:
            # ------------------------------------------------------------------
            # Create csv dataset from partitioned dataset for partitions
            # ------------------------------------------------------------------

            # partition loader
            evaluation_partitions_name = f"evaluation_partition_{Path(partition).stem}"
            input_dataset = CSVDataSet(filepath=partition)
            catalog.add(evaluation_partitions_name, input_dataset)

            # ------------------------------------------------------------------
            # Create txt dataset from partitioned dataset for qasm circuits
            # ------------------------------------------------------------------

            # qasm_circuits loader
            qasm_circuits = partition.replace(
                "02_intermediate", "03_qasm_circuits"
            ).replace(".csv", ".txt")
            qasm_circuits_name = f"qasm_circuit_{Path(qasm_circuits).stem}"
            input_dataset = TextDataSet(filepath=qasm_circuits)
            catalog.add(qasm_circuits_name, input_dataset)

            # ------------------------------------------------------------------
            # Create csv dataset from partitioned dataset for measures
            # ------------------------------------------------------------------

            # evaluation durations
            measures = partition.replace("02_intermediate", "04_measures")
            measures_name = f"measure_{Path(measures).stem}"
            input_dataset = CSVDataSet(filepath=measures)
            catalog.add(measures_name, input_dataset)

            # ------------------------------------------------------------------
            # Create csv dataset from partitioned dataset for evaluation durations
            # ------------------------------------------------------------------

            # evaluation durations
            execution_duration = partition.replace(
                "02_intermediate", "06_execution_durations"
            )
            execution_duration_name = (
                f"execution_duration_{Path(execution_duration).stem}"
            )
            input_dataset = CSVDataSet(filepath=execution_duration)
            catalog.add(execution_duration_name, input_dataset)

            # ------------------------------------------------------------------
            # Create csv dataset from partitioned dataset for evaluation results
            # ------------------------------------------------------------------

            # evaluation results
            execution_result = partition.replace(
                "02_intermediate", "05_execution_results"
            )
            execution_result_name = f"execution_result_{Path(execution_result).stem}"
            input_dataset = CSVDataSet(filepath=execution_result)
            catalog.add(execution_result_name, input_dataset)



"""
This is a boilerplate pipeline 'data_science'
generated using Kedro 0.18.3
"""

from kedro.pipeline import node, pipeline
from quafel.pipelines.data_generation.nodes import (
    extract_partition_data,
    generate_random_qasm_circuit_from_partition,
    calculate_measures,
    generate_evaluation_matrix,
    generate_evaluation_partitions,
    qasm_circuit_to_qiskit,
)


def create_pipeline(partitions, existing_circuits, existing_measures, **kwargs) -> dict:
    pl_generate_evaluation_partitions = pipeline(
        [
            node(
                func=generate_evaluation_matrix,
                inputs={
                    "min_qubits": "params:data_generation.min_qubits",
                    "max_qubits": "params:data_generation.max_qubits",
                    "qubits_increment": "params:data_generation.qubits_increment",
                    "qubits_type": "params:data_generation.qubits_type",
                    "min_depth": "params:data_generation.min_depth",
                    "max_depth": "params:data_generation.max_depth",
                    "depth_increment": "params:data_generation.depth_increment",
                    "depth_type": "params:data_generation.depth_type",
                    "min_shots": "params:data_generation.min_shots",
                    "max_shots": "params:data_generation.max_shots",
                    "max_shots": "params:data_generation.max_shots",
                    "shots_increment": "params:data_generation.shots_increment",
                    "shots_type": "params:data_generation.shots_type",
                    "frameworks": "params:data_generation.frameworks",
                },
                outputs={
                    "evaluation_matrix": "evaluation_matrix",
                },
            ),
            node(
                func=generate_evaluation_partitions,
                inputs={
                    "evaluation_matrix": "evaluation_matrix",
                    "skip_combinations": "params:data_generation.skip_combinations",
                },
                outputs={
                    "evaluation_partitions": "evaluation_partitions",
                },
            ),
        ],
        outputs={"evaluation_partitions": "evaluation_partitions"},
    )

    pl_generate_qasm_circuits = pipeline(
        [
            *[
                node(
                    func=generate_random_qasm_circuit_from_partition,
                    inputs={
                        "partition": f"evaluation_partition_{i}",
                        "seed": "params:data_generation.seed",
                    },
                    outputs={
                        "qasm_circuit": f"qasm_circuit_{i}",
                        "n_shots": f"n_shots_{i}",
                        "framework": f"framework_{i}",
                        "qiskit_circuit": f"qiskit_circuit_{i}",
                    },
                    tags=["dynamic"],
                    name=f"generate_random_qasm_circuit_from_partition_{i}",
                )
                for i in partitions
                if i not in existing_circuits
            ],
            *[
                node(
                    func=extract_partition_data,
                    inputs={
                        "partition": f"evaluation_partition_{i}",
                    },
                    outputs={
                        "qubits": f"qubits_{i}",
                        "depth": f"depth_{i}",
                        "n_shots": f"n_shots_{i}",
                        "framework": f"framework_{i}",
                    },
                    tags=["dynamic"],
                    name=f"extract_partition_data_{i}",
                )
                for i in partitions
                if i in existing_circuits
            ],
            *[
                node(
                    func=qasm_circuit_to_qiskit,
                    inputs={
                        "qasm_circuit": f"qasm_circuit_{i}",
                    },
                    outputs={
                        "qiskit_circuit": f"qiskit_circuit_{i}",
                    },
                    tags=["dynamic"],
                    name=f"extract_circuit_{i}",
                )
                for i in partitions
                if i in existing_circuits
            ],
            *[
                node(
                    func=calculate_measures,
                    inputs={
                        "circuit": f"qiskit_circuit_{i}",
                        "samples_per_parameter": "params:data_generation.samples_per_parameter",  # noqa
                        "haar_samples_per_qubit": "params:data_generation.haar_samples_per_qubit",  # noqa
                        "seed": "params:data_generation.seed",
                    },
                    outputs={
                        "measure": f"measure_{i}",
                    },
                    tags=["dynamic"],
                    name=f"calculate_measures_{i}",
                )
                for i in partitions
                if i not in existing_measures
            ],
        ],
        inputs={
            # note that this dataset is dynamically created in the hooks,
            # so it is not directly available in the catalog
            **{
                f"evaluation_partition_{i}": f"evaluation_partition_{i}"
                for i in partitions
            },
        },
        outputs={
            **{f"n_shots_{i}": f"n_shots_{i}" for i in partitions},
            **{f"framework_{i}": f"framework_{i}" for i in partitions},
            **{
                f"qasm_circuit_{i}": f"qasm_circuit_{i}"
                for i in partitions
                if i not in existing_circuits
            },
            **{
                f"measure_{i}": f"measure_{i}"
                for i in partitions
                if i not in existing_measures
            },
        },
    )

    return {
        "pl_generate_evaluation_partitions": pl_generate_evaluation_partitions,
        "pl_generate_qasm_circuits": pl_generate_qasm_circuits,
    }



"""``DaskRunner`` is an ``AbstractRunner`` implementation. It can be
used to distribute execution of ``Node``s in the ``Pipeline`` across
a Dask cluster, taking into account the inter-``Node`` dependencies.
"""
from collections import Counter
from itertools import chain
from typing import Any, Dict

from distributed import Client, as_completed, worker_client
from kedro.framework.hooks.manager import (
    _create_hook_manager,
    _register_hooks,
)
from kedro.framework.project import settings
from kedro.io import AbstractDataset, DataCatalog
from kedro.pipeline import Pipeline
from kedro.pipeline.node import Node
from kedro.runner import AbstractRunner, run_node
from pluggy import PluginManager


class _DaskDataSet(AbstractDataset):
    """``_DaskDataSet`` publishes/gets named datasets to/from the Dask
    scheduler."""

    def __init__(self, name: str):
        self._name = name

    def _load(self) -> Any:
        try:
            with worker_client() as client:
                return client.get_dataset(self._name)
        except ValueError:
            # Upon successfully executing the pipeline, the runner loads
            # free outputs on the scheduler (as opposed to on a worker).
            Client.current().get_dataset(self._name)

    def _save(self, data: Any) -> None:
        with worker_client() as client:
            client.publish_dataset(data, name=self._name, override=True)

    def _exists(self) -> bool:
        return self._name in Client.current().list_datasets()

    def _release(self) -> None:
        Client.current().unpublish_dataset(self._name)

    def _describe(self) -> Dict[str, Any]:
        return dict(name=self._name)


class DaskRunner(AbstractRunner):
    """``DaskRunner`` is an ``AbstractRunner`` implementation. It can be
    used to distribute execution of ``Node``s in the ``Pipeline`` across
    a Dask cluster, taking into account the inter-``Node`` dependencies.
    """

    def __init__(self, client_args: Dict[str, Any] = {}, is_async: bool = False):
        """Instantiates the runner by creating a ``distributed.Client``.

        Args:
            client_args: Arguments to pass to the ``distributed.Client``
                constructor.
            is_async: If True, the node inputs and outputs are loaded and saved
                asynchronously with threads. Defaults to False.
        """
        super().__init__(is_async=is_async)
        Client(**client_args)

    def __del__(self):
        Client.current().close()

    def create_default_data_set(self, ds_name: str) -> _DaskDataSet:
        """Factory method for creating the default dataset for the runner.

        Args:
            ds_name: Name of the missing dataset.

        Returns:
            An instance of ``_DaskDataSet`` to be used for all
            unregistered datasets.
        """
        return _DaskDataSet(ds_name)

    @staticmethod
    def _run_node(
        node: Node,
        catalog: DataCatalog,
        is_async: bool = False,
        session_id: str = None,
        *dependencies: Node,
    ) -> Node:
        """Run a single `Node` with inputs from and outputs to the `catalog`.

        Wraps ``run_node`` to accept the set of ``Node``s that this node
        depends on. When ``dependencies`` are futures, Dask ensures that
        the upstream node futures are completed before running ``node``.

        A ``PluginManager`` instance is created on each worker because the
        ``PluginManager`` can't be serialised.

        Args:
            node: The ``Node`` to run.
            catalog: A ``DataCatalog`` containing the node's inputs and outputs.
            is_async: If True, the node inputs and outputs are loaded and saved
                asynchronously with threads. Defaults to False.
            session_id: The session id of the pipeline run.
            dependencies: The upstream ``Node``s to allow Dask to handle
                dependency tracking. Their values are not actually used.

        Returns:
            The node argument.
        """
        hook_manager = _create_hook_manager()
        _register_hooks(hook_manager, settings.HOOKS)

        return run_node(node, catalog, hook_manager, is_async, session_id)

    def _run(
        self,
        pipeline: Pipeline,
        catalog: DataCatalog,
        hook_manager: PluginManager,
        session_id: str = None,
    ) -> None:
        nodes = pipeline.nodes
        load_counts = Counter(chain.from_iterable(n.inputs for n in nodes))
        node_dependencies = pipeline.node_dependencies
        node_futures = {}

        client = Client.current()
        for node in nodes:
            dependencies = (
                node_futures[dependency] for dependency in node_dependencies[node]
            )
            node_futures[node] = client.submit(
                DaskRunner._run_node,
                node,
                catalog,
                self._is_async,
                session_id,
                *dependencies,
            )

        for i, (_, node) in enumerate(
            as_completed(node_futures.values(), with_results=True)
        ):
            self._logger.info("Completed node: %s", node.name)
            self._logger.info("Completed %d out of %d tasks", i + 1, len(nodes))

            # Decrement load counts, and release any datasets we
            # have finished with. This is particularly important
            # for the shared, default datasets we created above.
            for data_set in node.inputs:
                load_counts[data_set] -= 1
                if load_counts[data_set] < 1 and data_set not in pipeline.inputs():
                    catalog.release(data_set)
            for data_set in node.outputs:
                if load_counts[data_set] < 1 and data_set not in pipeline.outputs():
                    catalog.release(data_set)

    def run_only_missing(
        self, pipeline: Pipeline, catalog: DataCatalog
    ) -> Dict[str, Any]:
        """Run only the missing outputs from the ``Pipeline`` using the
        datasets provided by ``catalog``, and save results back to the
        same objects.

        Args:
            pipeline: The ``Pipeline`` to run.
            catalog: The ``DataCatalog`` from which to fetch data.
        Raises:
            ValueError: Raised when ``Pipeline`` inputs cannot be
                satisfied.

        Returns:
            Any node outputs that cannot be processed by the
            ``DataCatalog``. These are returned in a dictionary, where
            the keys are defined by the node outputs.
        """
        free_outputs = pipeline.outputs() - set(catalog.list())
        missing = {ds for ds in catalog.list() if not catalog.exists(ds)}
        to_build = free_outputs | missing
        to_rerun = pipeline.only_nodes_with_outputs(*to_build) + pipeline.from_inputs(
            *to_build
        )

        # We also need any missing datasets that are required to run the
        # `to_rerun` pipeline, including any chains of missing datasets.
        unregistered_ds = pipeline.data_sets() - set(catalog.list())
        # Some of the unregistered datasets could have been published to
        # the scheduler in a previous run, so we need not recreate them.
        missing_unregistered_ds = {
            ds_name
            for ds_name in unregistered_ds
            if not self.create_default_data_set(ds_name).exists()
        }
        output_to_unregistered = pipeline.only_nodes_with_outputs(
            *missing_unregistered_ds
        )
        input_from_unregistered = to_rerun.inputs() & missing_unregistered_ds
        to_rerun += output_to_unregistered.to_outputs(*input_from_unregistered)

        # We need to add any previously-published, unregistered datasets
        # to the catalog passed to the `run` method, so that it does not
        # think that the `to_rerun` pipeline's inputs are not satisfied.
        catalog = catalog.shallow_copy()
        for ds_name in unregistered_ds - missing_unregistered_ds:
            catalog.add(ds_name, self.create_default_data_set(ds_name))

        return self.run(to_rerun, catalog)



from . import frameworks as fw
import time
import pandas as pd
from typing import Dict
import logging
import traceback
import re

log = logging.getLogger(__name__)


def measure_execution_durations(
    evaluations: int,
    **kwargs,
):
    ident = int(re.findall(r"\d+", list(kwargs.keys())[0])[-1])
    for key, value in kwargs.items():
        if key.startswith("framework_id"):
            framework_id = value
        elif key.startswith("qasm_circuit"):
            qasm_circuit = value
        elif key.startswith("n_shots"):
            n_shots = value

    try:
        framework = getattr(fw, framework_id)
    except AttributeError:
        raise AttributeError(
            f"Framework identifier does not match one of the existing frameworks.\
            Existing frameworks are {fw}"
        )

    framework_instance = framework(qasm_circuit=qasm_circuit, n_shots=int(n_shots))

    execution_perf_durations = []
    execution_proc_durations = []
    execution_results = []

    for e in range(evaluations):
        try:
            start_perf = time.perf_counter()
            start_proc = time.process_time()
            framework_instance.execute()
            finish_perf = time.perf_counter()
            finish_proc = time.process_time()
            execution_perf_durations.append(finish_perf - start_perf)
            execution_proc_durations.append(finish_proc - start_proc)
        except Exception:
            log.error(
                f"Error executing framework {framework_id} for experiment id {ident}: \
                    Execution failed in evaluation {e}: \
                        {traceback.format_exc()}"
            )
            # mark the whole set invalid
            execution_results = [0 for _ in range(evaluations)]
            break

        execution_results.append(framework_instance.get_result())

    return {
        "execution_duration": pd.DataFrame(
            {
                f"perf_{ident}": execution_perf_durations,
                f"proc_{ident}": execution_proc_durations,
            }
        ),
        "execution_result": pd.DataFrame({ident: execution_results}),
    }


def combine_evaluations(
    evaluation_partitions: Dict,
    execution_durations: Dict,
    execution_results: Dict,
    measures: Dict,
    export_results: bool,
):
    combine_all = pd.DataFrame()

    for (
        (partition_id, partition_load_func),
        (duration_id, duration_load_func),
        (result_id, result_load_func),
        (measure_id, measure_load_func),
    ) in zip(
        evaluation_partitions.items(),
        execution_durations.items(),
        execution_results.items(),
        measures.items(),
    ):
        assert (
            (partition_id == duration_id)
            and (partition_id == result_id)
            and (partition_id == measure_id)
        ), "Partition identifiers do not match duration and result identifiers."
        partition_data = partition_load_func()
        duration_data = duration_load_func()
        result_data = result_load_func() if export_results else None
        measure_data = measure_load_func()

        # TODO: unify somehow with the generation part
        partition_data.index = ["framework", "qubits", "depth", "shots"]

        duration_data_proc = duration_data.filter(regex="proc")
        duration_data_proc.index = [
            f"duration_proc_{i}" for i in range(len(duration_data))
        ]
        duration_data_proc.columns = [partition_id]

        duration_data_perf = duration_data.filter(regex="perf")
        duration_data_perf.index = [
            f"duration_perf_{i}" for i in range(len(duration_data))
        ]
        duration_data_perf.columns = [partition_id]

        duration_data = pd.concat([duration_data_perf, duration_data_proc])

        if result_data is not None:
            result_data.index = [f"result_{i}" for i in range(len(result_data))]

        measure_data = measure_data.T
        measure_data.columns = [partition_id]

        combined_partition_duration = pd.concat(
            (
                [partition_data, duration_data, result_data, measure_data]
                if result_data is not None
                else [partition_data, duration_data, measure_data]
            ),
            ignore_index=False,
            axis=0,
        )

        combine_all = pd.concat(
            [combine_all, combined_partition_duration],
            ignore_index=False,
            axis=1,
        )

    combine_all = combine_all.transpose()
    # combine_all = combine_all.sort_values(by=[0, 1, 3]) # sort by framework and shots

    return {"evaluations_combined": combine_all}



"""
This is a boilerplate pipeline 'visualization'
generated using Kedro 0.18.4
"""

import plotly.graph_objs as go
import plotly.express as px
import plotly.io as pio

from typing import Dict, List
import pandas as pd

from bisect import bisect_left
from math import log10, floor
import logging


import numpy as np

import os

duration_perf_regex = r"duration_perf_\d*"
duration_proc_regex = r"duration_proc_\d*"
expressibility_regex = r"expressibility"
entangling_capability = r"entangling_capability"


class design:
    # see https://plotly.com/python/discrete-color/
    qual_main = px.colors.qualitative.Dark2  # set1
    qual_second = px.colors.qualitative.Pastel2  # pastel1

    seq_main = px.colors.sequential.thermal  # alternative: pastel1

    print_figure_title = False

    title_font_size = 18
    legend_font_size = 16

    scatter_legend = dict(
        orientation="v",
        traceorder="normal",
        font=dict(
            size=legend_font_size,
        ),
    )

    base_theme = "simple_white"

    include_framework_term = (
        False  # Set to true to get e.g. 'Framework Pennylane' instead of 'Pennylane'
    )

    standard_ticks_angle = 0
    long_ticks_angle = -40
    long_ticks = 3

    showgrid = False

    time_tick_type = "log"
    time_dtick = 1

    log_tick_type = "log"
    standard_tick_type = "linear"

    heatmap_axis_mode = "array"
    scatter_axis_mode = "array"

    scatter_mode_c = "lines+markers"
    scatter_mode_hl = "lines"
    marker_color = dict(color="#444")


def rgb_to_rgba(rgb_value: str, alpha: float):
    """
    Adds the alpha channel to an RGB Value and returns it as an RGBA Value
    :param rgb_value: Input RGB Value
    :param alpha: Alpha Value to add  in range [0,1]
    :return: RGBA Value
    """
    return f"rgba{rgb_value[3:-1]}, {alpha})"


def get_time_scale(pd_time: pd.array):
    """This method takes an array of timestamps and finds the
    approriate time scale as well as the corresponding (exponential) factor.

    Args:
        pd_time (pd.array): pandas array of timestamps
    """

    def find_exp(number) -> int:
        """
        From https://stackoverflow.com/questions/64183806/extracting-the-exponent-from-scientific-notation # noqa
        """
        if number == 0:
            return 0
        base10 = log10(abs(number))
        return floor(base10)

    mid = np.mean(
        [
            find_exp(pd_time.min().min()),
            find_exp(pd_time.max().max()),
        ]
    ).astype(int)

    si_time = [["ns", "us", "ms", "s"], [-9, -6, -3, 0], [1e9, 1e6, 1e3, 1]]

    idx = bisect_left(si_time[1], mid)

    return (si_time[0][idx], si_time[2][idx])


def extract_framework_name_from_id(identifier):
    """
    Global method to generate display name from class name.
    E.g.: qiskit_fw -> Qiskit Framework
    """
    if design.include_framework_term:
        return identifier.replace("fw", "framework").capitalize().replace("_", " ")
    else:
        return identifier.replace("fw", "").capitalize().replace("_", " ")


def heatmap_viz(
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    z_title: str,
    x_title: str,
    log_x: bool,
    y_title: str,
    log_y: bool,
    plot_title: str,
):
    fig = go.Figure(
        [
            go.Heatmap(
                x=np.log2(x) if log_x else x,
                y=np.log2(y) if log_y else y,
                z=np.log10(z),
                colorscale=design.seq_main,
                colorbar=dict(
                    title=z_title,
                    tickvals=np.log10([z.min(), z.mean(), z.max()]),
                    ticktext=[f"{z.min():.2}", f"{z.mean():.2}", f"{z.max():.2}"],
                ),
            )
        ]
    )
    fig.update_layout(
        yaxis=dict(
            type="linear",
            tickmode=design.heatmap_axis_mode,
            tickvals=np.log2(y) if log_y else y,
            ticktext=y if log_y else None,
            tickangle=(
                design.long_ticks_angle
                if len(str(max(y))) >= design.long_ticks
                else design.standard_ticks_angle
            ),
            title=y_title,
            showgrid=design.showgrid,
        ),
        xaxis=dict(
            type="linear",
            tickmode=design.heatmap_axis_mode,
            tickvals=np.log2(x) if log_x else x,
            ticktext=x if log_x else None,
            tickangle=(
                design.long_ticks_angle
                if len(str(max(x))) >= design.long_ticks
                else design.standard_ticks_angle
            ),
            title=x_title,
            showgrid=design.showgrid,
        ),
        title=dict(
            text=plot_title if design.print_figure_title else "",
            font=dict(
                size=design.title_font_size,
            ),
        ),
        hovermode="x",
        font=dict(
            size=design.legend_font_size,
        ),
    )
    return fig


def scatter_viz(
    fig: go.Figure,
    name: str,
    main_color_sel: str,
    sec_color_sel: str,
    x: np.ndarray,
    y: np.ndarray,
    x_title: str,
    log_x: bool,
    y_title: str,
    log_y: bool,
    plot_title: str,
    y_max: np.ndarray = None,
    y_min: np.ndarray = None,
):
    fig.add_trace(
        go.Scatter(
            name=f"{name}",
            x=x,
            y=y,
            mode=design.scatter_mode_c,
            line=dict(color=main_color_sel),
        ),
    )
    if y_max is not None and y_min is not None:
        fig.add_trace(
            go.Scatter(
                name=f"{name} - High",
                x=x,
                y=y_max,
                mode=design.scatter_mode_hl,
                marker=design.marker_color,
                line=dict(width=0),
                showlegend=False,
            ),
        )
        fig.add_trace(
            go.Scatter(
                name=f"{name} - Low",
                x=x,
                y=y_min,
                marker=design.marker_color,
                line=dict(width=0),
                mode=design.scatter_mode_hl,
                fillcolor=sec_color_sel,
                fill="tonexty",
                showlegend=False,
            ),
        )
    fig.update_layout(
        xaxis=dict(
            type=design.log_tick_type if log_x else design.standard_tick_type,
            tickmode=design.scatter_axis_mode,
            tickvals=x,
            tickangle=(
                design.long_ticks_angle
                if len(str(max(x))) >= design.long_ticks
                else design.standard_ticks_angle
            ),
            title=x_title,
            showgrid=design.showgrid,
        ),
        yaxis=(
            dict(
                title=y_title,
                type=design.log_tick_type if log_y else None,
                dtick=design.time_dtick if log_y else None,
                showgrid=design.showgrid,
            )
        ),
        title=dict(
            text=plot_title if design.print_figure_title else "",
            font=dict(
                size=design.title_font_size,
            ),
        ),
        hovermode="x",
        font=dict(
            size=design.legend_font_size,
        ),
        legend=(design.scatter_legend),
        template=design.base_theme,
    )


def shots_qubits_viz(evaluations_combined: Dict):
    figures = {}

    si_time, factor_time = get_time_scale(
        evaluations_combined.filter(regex=duration_perf_regex)
    )

    grouped_by_fw = evaluations_combined.groupby("framework")

    for fw, qubit_depth_duration in grouped_by_fw:
        framework_name = extract_framework_name_from_id(fw)
        grouped_by_qubit = qubit_depth_duration.groupby("qubits")

        for q, depth_duration in grouped_by_qubit:
            duration_sorted_by_depth = depth_duration.sort_values("depth")
            durations = duration_sorted_by_depth.filter(regex=duration_perf_regex)

            durations *= factor_time

            duration_mean = durations.mean(axis=1)

            q = int(q)

            figures[f"{fw}_qubits_{q}_time"] = heatmap_viz(
                x=duration_sorted_by_depth["shots"].astype(int),
                y=duration_sorted_by_depth["depth"].astype(int),
                z=duration_mean,
                z_title=f"Time ({si_time})",
                x_title="# of Shots",
                log_x=True,
                y_title="Circuit Depth",
                log_y=True,
                plot_title=f"{framework_name} @ {q} Qubits: "
                f"Circuit Depth and # of Shots",
            )

    return figures


def shots_depths_viz(evaluations_combined: Dict):
    figures = {}

    si_time, factor_time = get_time_scale(
        evaluations_combined.filter(regex=duration_perf_regex)
    )

    grouped_by_fw = evaluations_combined.groupby("framework")

    for fw, qubit_depth_duration in grouped_by_fw:
        framework_name = extract_framework_name_from_id(fw)
        grouped_by_depth = qubit_depth_duration.groupby("depth")

        for d, qubit_duration in grouped_by_depth:
            duration_sorted_by_qubit = qubit_duration.sort_values("qubits")
            durations = duration_sorted_by_qubit.filter(regex=duration_perf_regex)

            durations *= factor_time

            duration_mean = durations.mean(axis=1)

            d = int(d)
            # image = []
            # for s, duration in grouped_by_shots_sorted_by_depth:
            #     image.append(duration['4'].to_numpy())

            figures[f"{fw}_depth_{d}_time"] = heatmap_viz(
                x=duration_sorted_by_qubit["shots"].astype(int),
                y=duration_sorted_by_qubit["qubits"].astype(int),
                z=duration_mean,
                z_title=f"Time ({si_time})",
                x_title="# of Shots",
                log_x=True,
                y_title="# of Qubits",
                log_y=False,
                plot_title=f"{framework_name} @ Circuit Depth {d}: "
                f"# of qubits and # of Shots",
            )

    return figures


def depth_qubits_viz(evaluations_combined: Dict):
    figures = {}

    si_time, factor_time = get_time_scale(
        evaluations_combined.filter(regex=duration_perf_regex)
    )

    grouped_by_fw = evaluations_combined.groupby("framework")

    for fw, qubit_depth_duration in grouped_by_fw:
        framework_name = extract_framework_name_from_id(fw)
        grouped_by_shots = qubit_depth_duration.groupby("shots")

        for s, shots_duration in grouped_by_shots:
            duration_sorted_by_depth = shots_duration.sort_values("depth")
            durations = duration_sorted_by_depth.filter(regex=duration_perf_regex)

            durations *= factor_time

            duration_mean = durations.mean(axis=1)

            s = int(s)

            figures[f"{fw}_shots_{s}_time"] = heatmap_viz(
                x=duration_sorted_by_depth["qubits"].astype(int),
                y=duration_sorted_by_depth["depth"].astype(int),
                z=duration_mean,
                z_title=f"Time ({si_time})",
                x_title="# of Qubits",
                log_x=False,
                y_title="Circuit Depth",
                log_y=True,
                plot_title=f"{framework_name} @ {s} Shots: "
                f"Circuit Depth and # of Qubits",
            )

    return figures


def qubits_time_viz(evaluations_combined: Dict, skip_frameworks: List):
    figures = {}

    # those two color sets are well suited as they correspond regarding
    # their color value but differ from their luminosity and saturation values
    main_colors_it = iter(design.qual_main)
    sec_colors_it = iter(design.qual_second)

    si_time, factor_time = get_time_scale(
        evaluations_combined.filter(regex=duration_perf_regex)
    )

    grouped_by_fw = evaluations_combined.groupby("framework")
    for fw, qubit_depth_duration in grouped_by_fw:
        if fw in skip_frameworks:
            continue

        main_color_sel = next(main_colors_it)
        sec_color_sel = rgb_to_rgba(next(sec_colors_it), 0.2)
        framework_name = extract_framework_name_from_id(fw)

        grouped_by_depth = qubit_depth_duration.groupby("depth")

        for d, qubit_shots_duration in grouped_by_depth:
            grouped_by_shots = qubit_shots_duration.groupby("shots")

            for s, fw_qubit_duration in grouped_by_shots:
                duration_sorted_by_qubit = fw_qubit_duration.sort_values("qubits")

                durations = duration_sorted_by_qubit.filter(regex=duration_perf_regex)
                durations *= factor_time

                durations_mean = durations.mean(axis=1)
                durations_max = durations.max(axis=1)
                durations_min = durations.min(axis=1)

                d = int(d)
                s = int(s)
                # image = []
                # for s, duration in grouped_by_shots_sorted_by_depth:
                #     image.append(duration['4'].to_numpy())

                if f"shots_{s}_depth_{d}_time" not in figures:
                    figures[f"shots_{s}_depth_{d}_time"] = go.Figure()

                scatter_viz(
                    fig=figures[f"shots_{s}_depth_{d}_time"],
                    name=f"{framework_name}",
                    main_color_sel=main_color_sel,
                    sec_color_sel=sec_color_sel,
                    x=duration_sorted_by_qubit["qubits"],
                    y=durations_mean,
                    y_min=durations_min,
                    y_max=durations_max,
                    x_title="# of Qubits",
                    log_x=False,
                    y_title=f"Time ({si_time})",
                    log_y=True,
                    plot_title=f"Duration per Framework over "
                    f"# of Qubits @ {s} Shots, Circuit Depth {d}",
                )

    return figures


def shots_time_viz(evaluations_combined: Dict, skip_frameworks: List):
    figures = {}

    # those two color sets are well suited as they correspond regarding
    # their color value but differ from their luminosity and saturation values
    main_colors_it = iter(design.qual_main)
    sec_colors_it = iter(design.qual_second)

    si_time, factor_time = get_time_scale(
        evaluations_combined.filter(regex=duration_perf_regex)
    )

    grouped_by_fw = evaluations_combined.groupby("framework")
    for fw, qubit_depth_duration in grouped_by_fw:
        if fw in skip_frameworks:
            continue

        main_color_sel = next(main_colors_it)
        sec_color_sel = rgb_to_rgba(next(sec_colors_it), 0.2)
        framework_name = extract_framework_name_from_id(fw)

        grouped_by_depth = qubit_depth_duration.groupby("depth")

        for d, qubit_shots_duration in grouped_by_depth:
            grouped_by_qubits = qubit_shots_duration.groupby("qubits")

            for q, fw_shots_duration in grouped_by_qubits:
                duration_sorted_by_shots = fw_shots_duration.sort_values("shots")

                durations = duration_sorted_by_shots.filter(regex=duration_perf_regex)
                durations *= factor_time

                durations_mean = durations.mean(axis=1)
                durations_max = durations.max(axis=1)
                durations_min = durations.min(axis=1)

                q = int(q)
                d = int(d)
                # image = []
                # for s, duration in grouped_by_shots_sorted_by_depth:
                #     image.append(duration['4'].to_numpy())

                if f"qubits_{q}_depth_{d}_time" not in figures:
                    figures[f"qubits_{q}_depth_{d}_time"] = go.Figure()

                scatter_viz(
                    fig=figures[f"qubits_{q}_depth_{d}_time"],
                    name=f"{framework_name}",
                    main_color_sel=main_color_sel,
                    sec_color_sel=sec_color_sel,
                    x=duration_sorted_by_shots["shots"].astype(int),
                    y=durations_mean,
                    y_min=durations_min,
                    y_max=durations_max,
                    x_title="# of Shots",
                    log_x=True,
                    y_title=f"Time ({si_time})",
                    log_y=True,
                    plot_title=f"Duration per Framework over "
                    f"# of Shots @ {q} Qubits, Circuit Depth {d}",
                )

    return figures


def depth_time_viz(evaluations_combined: Dict, skip_frameworks: List):
    figures = {}

    # those two color sets are well suited as they correspond regarding
    # their color value but differ from their luminosity and saturation values
    main_colors_it = iter(design.qual_main)
    sec_colors_it = iter(design.qual_second)

    si_time, factor_time = get_time_scale(
        evaluations_combined.filter(regex=duration_perf_regex)
    )

    grouped_by_fw = evaluations_combined.groupby("framework")
    for fw, qubit_depth_duration in grouped_by_fw:
        if fw in skip_frameworks:
            continue

        main_color_sel = next(main_colors_it)
        sec_color_sel = rgb_to_rgba(next(sec_colors_it), 0.2)
        framework_name = extract_framework_name_from_id(fw)

        grouped_by_qubits = qubit_depth_duration.groupby("qubits")

        for q, depth_shots_duration in grouped_by_qubits:
            grouped_by_shots = depth_shots_duration.groupby("shots")

            for s, fw_depth_duration in grouped_by_shots:
                duration_sorted_by_depth = fw_depth_duration.sort_values("depth")

                durations = duration_sorted_by_depth.filter(regex=duration_perf_regex)
                durations *= factor_time

                durations_mean = durations.mean(axis=1)
                durations_max = durations.max(axis=1)
                durations_min = durations.min(axis=1)

                q = int(q)
                s = int(s)
                # image = []
                # for s, duration in grouped_by_shots_sorted_by_depth:
                #     image.append(duration['4'].to_numpy())

                if f"shots_{s}_qubits_{q}_time" not in figures:
                    figures[f"shots_{s}_qubits_{q}_time"] = go.Figure()

                scatter_viz(
                    fig=figures[f"shots_{s}_qubits_{q}_time"],
                    name=f"{framework_name}",
                    main_color_sel=main_color_sel,
                    sec_color_sel=sec_color_sel,
                    x=duration_sorted_by_depth["depth"].astype(int),
                    y=durations_mean,
                    y_min=durations_min,
                    y_max=durations_max,
                    x_title="Circuit Depth",
                    log_x=True,
                    y_title=f"Time ({si_time})",
                    log_y=True,
                    plot_title=f"Duration per Framework over "
                    f"Circuit Depth @ {s} Shots, {q} Qubits",
                )

    return figures


def qubits_measures_viz(evaluations_combined: Dict) -> Dict[str, go.Figure]:
    """Returns a dictionary of subplots, where the keys are of the form
    'shots_<n>_depth_<m>_measures', where <n> and <m> are the shot number and
    circuit depth, respectively. Each subplot shows the mean expressibility
    and mean entangling capability (over all qubits) of each framework for a
    given shot number and circuit depth.

    Args:
        evaluations_combined (Dict): A dataframe containing the evaluations.

    Returns:
        Dict[str, Figure]: A dictionary of subplots, where the keys are of the
            form 'shots_<n>_depth_<m>_measures', and the values are subplots
            objects.
    """

    figures = {}

    grouped_by_depth = evaluations_combined.groupby("depth")

    for d, qubit_shots_duration in grouped_by_depth:
        grouped_by_shots = qubit_shots_duration.groupby("shots")

        for s, fw_qubit_duration in grouped_by_shots:
            duration_sorted_by_qubit = fw_qubit_duration.sort_values("qubits")

            expressiblities = duration_sorted_by_qubit.filter(
                regex=expressibility_regex
            ).mean(axis=1)
            entangling_capabilities = duration_sorted_by_qubit.filter(
                regex=entangling_capability
            ).mean(axis=1)

            d = int(d)
            s = int(s)

            figures[f"shots_{s}_depth_{d}_measures"] = go.Figure()

            scatter_viz(
                fig=figures[f"shots_{s}_depth_{d}_measures"],
                name="Entangling Capability",
                main_color_sel=design.qual_main[0],
                sec_color_sel=rgb_to_rgba(design.qual_second[0], 0.2),
                x=duration_sorted_by_qubit["qubits"],
                y=entangling_capabilities,
                x_title="# of Qubits",
                log_x=False,
                y_title="Measures",
                log_y=False,
                plot_title=f"Measures per Framework over "
                f"# of Qubits @ {s} Shots, Circuit Depth {d}",
            )

            scatter_viz(
                fig=figures[f"shots_{s}_depth_{d}_measures"],
                name="Expressibility",
                main_color_sel=design.qual_main[1],
                sec_color_sel=rgb_to_rgba(design.qual_second[1], 0.2),
                x=duration_sorted_by_qubit["qubits"],
                y=expressiblities,
                x_title="# of Qubits",
                log_x=False,
                y_title="Measures",
                log_y=False,
                plot_title=f"Measures per Framework over "
                f"# of Qubits @ {s} Shots, Circuit Depth {d}",
            )

    return figures


def shots_measures_viz(evaluations_combined: Dict):
    """
    Visualizes measures based on shots and depth.

    Parameters:
    - evaluations_combined: a dictionary containing evaluations

    Returns:
    - figures: a dictionary containing the visualizations
    """
    figures = {}

    grouped_by_depth = evaluations_combined.groupby("depth")

    for d, qubit_shots_duration in grouped_by_depth:
        grouped_by_qubits = qubit_shots_duration.groupby("qubits")

        for q, fw_shots_duration in grouped_by_qubits:
            duration_sorted_by_shots = fw_shots_duration.sort_values("shots")

            expressiblities = duration_sorted_by_shots.filter(
                regex=expressibility_regex
            ).mean(axis=1)
            entangling_capabilities = duration_sorted_by_shots.filter(
                regex=entangling_capability
            ).mean(axis=1)

            q = int(q)
            d = int(d)

            figures[f"qubits_{q}_depth_{d}_measures"] = go.Figure()

            scatter_viz(
                fig=figures[f"qubits_{q}_depth_{d}_measures"],
                name="Entangling Capability",
                main_color_sel=design.qual_main[0],
                sec_color_sel=rgb_to_rgba(design.qual_second[0], 0.2),
                x=duration_sorted_by_shots["shots"].astype(int),
                y=entangling_capabilities,
                x_title="# of Shots",
                log_x=True,
                y_title="Measures",
                log_y=False,
                plot_title=f"Measures per Framework over "
                f"# of Shots @ {q} Qubits, Circuit Depth {d}",
            )

            scatter_viz(
                fig=figures[f"qubits_{q}_depth_{d}_measures"],
                name="Expressibility",
                main_color_sel=design.qual_main[1],
                sec_color_sel=rgb_to_rgba(design.qual_second[1], 0.2),
                x=duration_sorted_by_shots["shots"].astype(int),
                y=expressiblities,
                x_title="# of Shots",
                log_x=True,
                y_title="Measures",
                log_y=False,
                plot_title=f"Measures per Framework over "
                f"# of Shots @ {q} Qubits, Circuit Depth {d}",
            )

    return figures


def depth_measures_viz(evaluations_combined: Dict):
    """
    Generates visualizations of measures for different qubits and
    shots based on the provided evaluations.

    Parameters:
    - evaluations_combined (Dict): A dictionary containing
    evaluations combined by qubits and shots.

    Returns:
    - figures (Dict): A dictionary containing visualizations of
    measures for different qubits and shots.
    """
    figures = {}

    grouped_by_qubits = evaluations_combined.groupby("qubits")

    for q, depth_shots_duration in grouped_by_qubits:
        grouped_by_shots = depth_shots_duration.groupby("shots")

        for s, fw_depth_duration in grouped_by_shots:
            duration_sorted_by_depth = fw_depth_duration.sort_values("depth")

            expressiblities = duration_sorted_by_depth.filter(
                regex=expressibility_regex
            ).mean(axis=1)
            entangling_capabilities = duration_sorted_by_depth.filter(
                regex=entangling_capability
            ).mean(axis=1)

            q = int(q)
            s = int(s)

            figures[f"shots_{s}_qubits_{q}_measures"] = go.Figure()

            scatter_viz(
                fig=figures[f"shots_{s}_qubits_{q}_measures"],
                name="Entangling Capability",
                main_color_sel=design.qual_main[0],
                sec_color_sel=rgb_to_rgba(design.qual_second[0], 0.2),
                x=duration_sorted_by_depth["depth"].astype(int),
                y=entangling_capabilities,
                x_title="Circuit Depth",
                log_x=True,
                y_title="Measures",
                log_y=False,
                plot_title=f"Measures per Framework over "
                f"Circuit Depth @ {s} Shots, {q} Qubits",
            )

            scatter_viz(
                fig=figures[f"shots_{s}_qubits_{q}_measures"],
                name="Expressibility",
                main_color_sel=design.qual_main[1],
                sec_color_sel=rgb_to_rgba(design.qual_second[1], 0.2),
                x=duration_sorted_by_depth["depth"].astype(int),
                y=expressiblities,
                x_title="Circuit Depth",
                log_x=True,
                y_title="Measures",
                log_y=False,
                plot_title=f"Measures per Framework over "
                f"Circuit Depth @ {s} Shots, {q} Qubits",
            )

    return figures


def extract_tests(evaluations_combined: Dict):
    si_time, factor_time = get_time_scale(
        evaluations_combined.filter(regex=duration_perf_regex)
    )

    durations_perf = evaluations_combined.filter(regex=duration_perf_regex)

    durations_proc = evaluations_combined.filter(regex=duration_proc_regex)

    n_runs = durations_perf.shape[0]
    n_evals = durations_perf.shape[1]

    evals_perf_mean = durations_perf.mean(axis=0)
    evals_proc_mean = durations_proc.mean(axis=0)

    log = logging.getLogger(__name__)
    log.info(f"Received {n_evals} evaluations and {n_runs} runs")
    log.info(
        f"Evaluation mean across all runs (perf timing): \
            {evals_perf_mean.mean()}"
    )
    log.info(
        f"Evaluation deviation across all runs (perf timing): \
            {evals_perf_mean.max() - evals_perf_mean.min()}"
    )
    log.info(
        f"Evaluation mean across all runs (proc timing): \
             {evals_proc_mean.mean()}"
    )
    log.info(
        f"Evaluation deviation across all runs (proc timing):\
            {evals_proc_mean.max() - evals_proc_mean.min()}"
    )

    return {}


def export_selected(evaluations_combined, additional_figures, output_folder, **figures):
    max_qubits = evaluations_combined["qubits"].max()
    max_depth = evaluations_combined["depth"].max()
    max_shots = evaluations_combined["shots"].max()
    frameworks = evaluations_combined["framework"].unique()

    def export(fig, name, folder):
        # Disable warnings to prevent printing a box at the bottom left of the figure.
        # See this issue: https://github.com/plotly/plotly.py/issues/3469
        pio.full_figure_for_development(fig, warn=False)
        # Disable mathjax completely since above did not help.
        # See this issue: https://github.com/plotly/Kaleido/issues/122
        pio.kaleido.scope.mathjax = None

        fig.write_image(os.path.join(folder, f"{name}.pdf"), engine="kaleido")
        # set scale=3 to increase resolution in resulting plots
        fig.write_image(os.path.join(folder, f"{name}.png"), engine="kaleido", scale=3)

    sel = f"shots_{max_shots}_depth_{max_depth}_time"
    export(figures[sel], sel, output_folder)

    sel = f"shots_{max_shots}_qubits_{max_qubits}_time"
    export(figures[sel], sel, output_folder)

    sel = f"qubits_{max_qubits}_depth_{max_depth}_time"
    export(figures[sel], sel, output_folder)

    sel = f"shots_{max_shots}_depth_{max_depth}_measures"
    export(figures[sel], sel, output_folder)

    sel = f"shots_{max_shots}_qubits_{max_qubits}_measures"
    export(figures[sel], sel, output_folder)

    sel = f"qubits_{max_qubits}_depth_{max_depth}_measures"
    export(figures[sel], sel, output_folder)

    for fw in frameworks:
        sel = f"{fw}_qubits_{max_qubits}_time"
        export(figures[sel], sel, output_folder)

        sel = f"{fw}_depth_{max_depth}_time"
        export(figures[sel], sel, output_folder)

        sel = f"{fw}_shots_{max_shots}_time"
        export(figures[sel], sel, output_folder)

    for add_fig in additional_figures:
        export(figures[add_fig], add_fig, output_folder)

    return {}
